import json
import time
import string
from re import sub

from langchain_core.prompts import PromptTemplate
from langchain_community.callbacks.manager import get_openai_callback

from . import agent_events
from llmsdk.lib import defaults
from llmsdk.lib.enablers import AgentEnablersMixin
from llmsdk.agents.basellmrag import BaseLLMRAGAgent
from llmsdk.lib import SafeEncoder

__all__ = ['LLMDatabankQuerier']
DEFAULT_TOPK = 10

class LLMDatabankQuerier(BaseLLMRAGAgent, AgentEnablersMixin):
    """
    Class to load and index a docset and support running queries against the docset
    for retrieval of most similar matches
    """

    def __init__(self,
                 name,
                 cred={},
                 platform=defaults.LLM_PLATFORM,
                 model=defaults.LLM_MODEL,
                 embedding_model=defaults.LLM_EMBEDDING_MODEL,
                 statestore=defaults.LLM_STATE_STORE):
        """
        init the LLM query agent
        name: name of the agent
        cred: credentials object
        platform: name of the LLM platform backend to use
                default to OpenAI GPT platform for now, Azure is also supported
                will be extended in the future to suuport other models
        memory_size: how many tokens of memory to use when chatting with the LLM
        """

        start_time = time.time()

        # init the base class
        super().__init__(name=name,
                         cred=cred,
                         platform=platform,
                         model=model,
                         embedding_model=embedding_model,
                         agent_type="databank-query",
                         statestore=statestore)

        # defaults
        self.chunk_size = 1000
        self.chunk_overlap = 300
        self.index = None
        self.metadata = {}
        self.vdb_client = None
        self.index_name = None
        self.index_store = None
        self.docs = {}

        # LLM params
        self.platform = platform
        self.model = model
        self.embedding_model = embedding_model

        # init the llm and embeddings objects
        self.llm, self.embeddings = self._get_llm_objs(platform=self.platform,
                                                        model=self.model,
                                                        embedding_model=self.embedding_model,
                                                        cred=self.cred)

        # note metadata for this agent
        self.metadata = {
            "agent": {
                "name": self.agent_name,
                "type": self.agent_type,
                "platform": self.platform,
                "model": self.model,
                "embedding_model": self.embedding_model,
            },
            "events": []
        }
        # log that the agent is ready
        duration = time.time() - start_time
        event = self._log_event(agent_events._EVNT_READY, duration)

    ## helper functions

    def run_query(self, query, filters, topk):
        """
        run a query against an internal docset indexed by the agent
        """

        # get the similar docs
        docs = self.get_similar_docs(query, filters=filters, topk=topk)

        ##
        ## -- TO-DO --
        ## can add more complex processing of returned docs here
        ## e.g. re-ranking, summarization by LLM, etc.
        ## -- END TO-DO --
        ##

        # construct the result
        result = {
            "question": query,
            "sources": [{
                "content": d.page_content,
                "metadata": d.metadata,
                "distance": d.metadata.pop('distance')
            } for d in docs],
        }

        return result

    def query(self, query, filters=None, topk=DEFAULT_TOPK):
        """
        run a query on an index
        query: query string
        topk: number of closest matching docs to return
        """

        start_time = time.time()

        try:
            if self.platform in ['openai', 'azure']:
                with get_openai_callback() as cb:
                    result = self.run_query(query, filters=filters, topk=topk)
                stats = {
                    "total_tokens": cb.total_tokens,
                    "prompt_tokens": cb.prompt_tokens,
                    "completion_tokens": cb.completion_tokens,
                    "total_cost": round(cb.total_cost, 4)
                }
            else:
                result = self.run_query(query, filters=filters, topk=topk)
                stats = {}
        except:
            result = {
                "question": query,
                "sources": [],
            }
            stats = {}

        # log the event
        params = {
            "query": query,
            "topk": topk,
            "result": result.copy() if result is not None else None,
            "stats": stats,
        }
        duration = time.time() - start_time
        event = self._log_event(agent_events._EVNT_QUERY, duration, params=params)

        # add the event to the result
        result['metadata'] = {
            "timestamp": event['timestamp'],
            "duration": event['duration'],
        }

        return result


################### BEGIN TESTBED ###################

if __name__ == "__main__":

    # vars
    # cred = get_credentials_by_name('openai-api')
    persist_directory = "chromadb"
    path        = '...'

    agentname   = "broker-querier"
    platform    = "openai"
    model       = "gpt-4o-mini"
    mode        = "load"

    ## create an agent
    # agent to do databank querying
    agent = LLMDatabankQuerier(name=agentname, platform=platform, model=model)

    if mode == "create":
        ## load the data
        # set the params
        # 'index_columns' specifies names of columns in dataset that will be vector indexed
        params = {
            "index_columns": ["name", "summary"]
        }
        # specify source='csv' for CSV datasets
        data = agent.load_data(content=path,
                                source='csv',
                                params=params)

        ## index the data
        # create the index and add data to it
        # this call will flush an existing index if it already exists
        # can be called multiple times
        agent.create_add_index(data=data,
                               store="chroma",
                               persist_directory=persist_directory,
                               index_name=agentname)
    else:
        agent.load_index(persist_directory=persist_directory,
                            index_name=agentname,
                            store='chroma')

    ## run the query
    query = "platform used by MetLife for group insurance operations"
    # set the filters if needed
    # '_fieldtype' is a list containing the fields on which search should be executed
    #       if empty, then search is done across all indexed fields in the databnk
    filters = {
        "_fieldtype": ["name", "summary"]
    }
    # 'topk' specifies the number of results to return from the databank
    result = agent.query(query,
                         filters=filters,
                         topk=3)

    ## display
    print(json.dumps(result, indent=2, cls=SafeEncoder))



import os
import csv
import json
import time
import codecs
import pickle
import hashlib
import logging
from logging.config import dictConfig

from langchain import hub
from langchain_openai import ChatOpenAI
from langchain_openai import AzureChatOpenAI
from langchain_community.utilities import SerpAPIWrapper
from langchain.agents import create_structured_chat_agent
from langchain.agents import AgentExecutor
from langchain_core.tools import Tool
from langchain_core.documents import Document
from langchain_text_splitters import RecursiveCharacterTextSplitter
from langchain_community.vectorstores import Chroma
from langchain_community.vectorstores import FAISS
from langchain_community.document_loaders import Docx2txtLoader
from langchain_community.document_loaders import PyPDFLoader
from langchain_community.document_loaders import PyMuPDFLoader
from langchain_community.document_loaders import PDFMinerLoader
from langchain_community.document_loaders import UnstructuredHTMLLoader
from langchain_community.document_loaders import DirectoryLoader

import chromadb
from chromadb.config import Settings
from chromadb.utils.embedding_functions import OpenAIEmbeddingFunction

import fitz

from llmsdk.agents import agent_events
from llmsdk.lib import defaults
from llmsdk.lib.statemgmt import RedisMixin
from llmsdk.lib.enablers import AgentEnablersMixin
from llmsdk.lib import SafeEncoder

try:
    from llmsdk.services.log import *
    logging.config.dictConfig(log_config)
except:
    logging.basicConfig()

class BaseLLMRAGAgent(RedisMixin, AgentEnablersMixin):
    """
    Class to define the base Hasper LLM agent
    """

    def __init__(self,
                 name,
                 cred,
                 agent_type,
                 platform=defaults.LLM_PLATFORM,
                 model=defaults.LLM_MODEL,
                 embedding_model=defaults.LLM_EMBEDDING_MODEL,
                 searchapi=defaults.LLM_SEARCH_API,
                 statestore=defaults.LLM_STATE_STORE,
                 memory_size=1000,
                 topk=7):
        """
        init the LLM query agent
        name: name of the agent
        cred: credentials object
        platform: name of the LLM platform backend to use
                default to OpenAI GPT platform for now, Azure is also supported
                will be extended in the future to suuport other models
        memory_size: how many tokens of memory to use when chatting with the LLM
        """

        # logging
        self.logger = get_logger()

        # name
        start_time      = time.time()
        self.agent_name = name
        self.agent_type = agent_type
        self.agent_id   = self._create_id(f"{self.agent_name}_{start_time}")

        # How many documents/chunks to respond to?
        self.topk = topk

        # statekey of the agent
        # the statekey is a pointer to the current state of the agent
        # this state can be persisted to some store
        # when needed, any previous persisted state
        # can be recalled and reassigned to the agent
        # for now, statekey is static
        # in the future, we can store and retrieve possibly multiple states
        self.statekey = f"sk:{self.agent_type}#{self.agent_name}"

        # creds
        self.cred = cred

        # LLM params
        self.platform           = platform
        self.model              = model
        self.embedding_model    = embedding_model

        # init the llm and embeddings objects
        self.llm, self.embeddings = self._get_llm_objs(platform=self.platform,
                                                       model=self.model,
                                                       embedding_model=self.embedding_model,
                                                       cred=self.cred)

    ##
    ## interfaces
    ##

    def get_metadata(self):
        """
        return metadata collected by the agent
        """
        return self.metadata


    ## state management

    def get_statekey(self):
        """
        return the statekey which points to the
        latest state of the agent
        """
        return self.statekey

    def clear_agent_state(self):
        """
        clear the memory context of the agent
        this is useful when we want to agent to
        start off a new QnA session
        """
        start_time = time.time()

        # clear the memories
        self.memory.clear()
        self.kg.clear()

        # reset the context and KG vars
        self.latest_context = []
        self.current_kg = []

        # make note of the clear action
        duration = time.time() - start_time
        params = {}
        event = self._log_event(agent_events._EVNT_STATECLEAR, duration, params=params)

        return

    def get_agent_state(self):
        """
        get the state of the agent
        """
        # get the memories
        memory = self.memory
        kg = self.kg

        # get the context and KG vars
        latest_context = self.latest_context
        current_kg = self.current_kg

        # setup the state object
        state = {
            "memory": memory,
            "kg": kg,
            "latest_context": latest_context,
            "current_kg": current_kg,
        }

        return state

    def store_agent_state(self):
        """
        store the agent state in some external storage
        this is useful when we want to later retrieve
        the state and re-animate an agent with the retrieved state
        """
        start_time = time.time()

        # get the state
        state = self.get_agent_state()

        # pickle it for transport
        state_pkl = codecs.encode(pickle.dumps(state), "base64").decode()

        # store it
        success = False
        statekey = self.statekey
        if self.state_store == "redis":
            cred = self.cred.get(self.state_store, {})
            success = self.store_state_redis(statekey, state_pkl, cred)
        elif self.state_store == "db":
            pass
        elif self.state_store == "disk":
            pass
        else:
            raise Exception(f"Unknown state store for agent, cannot store the agent's state for {statekey}")

        # make note of the action
        if success:
            duration = time.time() - start_time
            params = {
                "statekey": statekey,
                "store": self.state_store
            }
            event = self._log_event(agent_events._EVNT_STATESTORE, duration, params=params)

        return success

    def set_agent_state(self, statekey):
        """
        get a previously stored state
        and re-animate the agent with it
        """
        start_time = time.time()
        success = False

        # retrieve the state from storage
        state_pkl = None
        if self.state_store == "redis":
            cred = self.cred.get(self.state_store, {})
            state_pkl = self.retrieve_state_redis(statekey, cred)
        elif self.state_store == "db":
            pass
        elif self.state_store == "disk":
            pass
        else:
            raise Exception(f"Unknown state store for agent, cannot retrieve state for {statekey}")

        state = None
        if state_pkl:
            # re-animate our agent with the retrieved state
            # unpickle the state pickle
            try:
                state = pickle.loads(codecs.decode(state_pkl.encode(), "base64"))
            except:
                self.logger.error("Cannot re-animate agent with retrieved state, setting clear state",
                                     extra={
                                         'source': self.agent_name,
                                         'data': f"State pickle: {state_pkl}"
                                     })
                # clear the agent state
                self.clear_agent_state()

        if state:
            # re-hydrate state vars
            if state.get('memory'):
                self.memory = state['memory']
            if state.get('kg'):
                self.kg = state['kg']
            if state.get('latest_context'):
                self.latest_context = state['latest_context']
            if state.get('current_kg'):
                self.current_kg = state['current_kg']
            success = True

        # note the action
        duration = time.time() - start_time
        params = {
            "statekey": statekey,
            "success": success,
            "store": self.state_store
        }
        event = self._log_event(agent_events._EVNT_STATERESTORE, duration, params=params)

        return


    ## data loading

    def load_csv_data(self, path, params):
        """
        custom loader for CSV data
        """

        self.logger.debug("Processing CSV data in file")

        # init
        data = []

        # get the list of columns to include while indexing
        index_columns = params.get('index_columns', [])
        if len(index_columns) == 0:
            return None

        # process the CSV
        record_cnt = 0
        with open(path, newline='') as csvfile:
            reader = csv.DictReader(csvfile)

            # for each row in the file
            for row in reader:
                metadata = row.copy()

                # iterate over all the fields
                for field in index_columns:
                    md = metadata.copy()
                    md['_row'] = record_cnt
                    # increment the record counter
                    record_cnt += 1

                    # tag the metadata with the field type
                    md['_fieldtype'] = field

                    # create the Document object
                    page_content = f"{row[field]}"
                    doc = Document(page_content=page_content, metadata=md)

                    # collect it
                    data.append(doc)


        # check for success
        if len(data) == 0:
            data = None

        return data

    def load_json_data(self, data, params):

        self.logger.debug("Processing JSON data")

        if isinstance(data, str):
            if os.path.exists(data):
                data = json.load(open(data))
            else:
                self.logger.error("Input is neither a file path, nor in-memory data")
                return None

        # get the list of columns to include while indexing
        index_columns = params.get('index_columns', [])
        if len(index_columns) == 0:
            self.logger.debug("Index columns not specified")
            return None

        if not isinstance(data, list) or len(data) == 0:
            self.logger.debug("Cant load data - not a valid list")
            return None

        docs = []
        invalid_cnt = 0
        missing_field_cnt = 0
        record_cnt = 0
        for row in data:
            if not isinstance(row, dict):
                invalid_cnt += 1
                continue

            for field in index_columns:
                if field not in row:
                    missing_field_cnt += 1
                    continue
                md = row.copy()
                md['_row'] = record_cnt
                record_cnt += 1

                # tag the metadata with the field type
                md['_fieldtype'] = field

                # create the Document object
                page_content = row[field]
                doc = Document(page_content=page_content, metadata=md)

                # collect it
                docs.append(doc)

        self.logger.debug(f"Loaded docs: {record_cnt} (invalid: {invalid_cnt} Missing index field: {missing_field_cnt}")

        # check for success
        if len(docs) == 0:
            docs = None

        return docs

    def chunk_data(self, data, _format):
        """
        create chunks from the data
        and add any needed metadata
        """

        def cleanup_metadata(data):
            # take in a list of data document objects
            # and clean up metadata
            curr_source = None
            for i in range(0, len(data)):

                source = data[i].metadata['source']
                content = data[i].page_content

                # add bboxes for pdf docs
                if _format == "pdf":
                    pageno = data[i].metadata['page']
                    page = self.docs[source].load_page(pageno)
                    instances = page.search_for(content)
                    bboxes = []
                    for inst in instances:
                        bboxes.append([inst.x0, inst.y0,
                                       inst.x1, inst.y1])
                    data[i].metadata['bboxes'] = json.dumps(bboxes)

                # add other metadata
                data[i].metadata['file'] = source.split('/')[-1]
                if curr_source != source:
                    curr_source = source
                    chunk = 1
                data[i].metadata['chunk'] = chunk
                if _format == "str":
                    data[i].metadata['id'] = self._create_id(f"{source}-{self._create_id(content)}-{chunk}")
                else:
                    data[i].metadata['id'] = self._create_id(f"{source}-{chunk}")
                chunk += 1

                ##
                ## add any other custom metadata here
                ##

            return data

        # chunk the data
        text_splitter = RecursiveCharacterTextSplitter(chunk_size=self.chunk_size,
                                                       chunk_overlap=self.chunk_overlap)
        chunks = text_splitter.split_documents(data)

        # add metadata
        chunks = cleanup_metadata(chunks)

        return chunks


    def load_data(self, content, source=None, metadata={}, params={}):
        """
        set the datasource loader, loads and cleans the data
        content: data content (path, text) depending on type of source
        source:
            'dir' points to a folder with one or more files to load
            'pdf' points to a single PDF file to load
            'csv' points to a single CSV file to load
            'json' points to JSON data in file or in-memory list to load
            'str' contains a text string to load
            None will try to infer automatically
        params: extra params needed for specific loaders
                    glob: what glob filter to use if source=='dir'
                    pdfloader: what type of pdf loader module to use if source=='pdf'
                    index_columns: what columns to index if source=='csv'
        metadata: any custom metadata to pass when source=='str'
        """

        start_time = time.time()
        success = False

        if not source:
            source = self._get_path_source(content)

        if source == 'dir':
            glob = params.get("glob", "**/*.*")
            loader = DirectoryLoader(content, glob=glob, recursive=True)
            data = loader.load()

        elif source == 'pdf':
            pdfloader = params.get("pdfloader", "pymupdf")
            if pdfloader == "pymupdf":
                loader = PyMuPDFLoader(content)
                data = loader.load()
            elif pdfloader == "pypdf":
                loader = PyPDFLoader(content)
                data = loader.load_and_split()
            elif pdfloader == "pypdfium2":
                loader = PyPDFium2Loader(content)
                data = loader.load()
            elif pdfloader == "pdfminer":
                loader = PDFMinerLoader(content)
                data = loader.load()
            else:
                data = None

            # make note of this document
            # we'll need to access it later for metadata
            if data:
                self.docs[content] = fitz.open(content)

        elif source == 'docx':
            loader = Docx2txtLoader(content)
            data = loader.load()

        elif source == 'csv':
            data = self.load_csv_data(content, params)
        elif source == 'json':
            data = self.load_json_data(content, params)

        elif source == 'str':
            # special handling for string inputs
            metadata['source'] = source
            data = [Document(page_content=content, metadata=metadata)]

        else:
            data = None

        # chunk the data
        if data:
            if source not in ['csv']:
                data = self.chunk_data(data, source)
            success = True

        # log that the data loader is ready
        duration = time.time() - start_time
        params = {
            "success": success,
            "source": source,
            "content": content,
            "params": params,
            "metadata": metadata,
        }
        event = self._log_event(agent_events._EVNT_DATA, duration, params=params)

        return data


    ## index management

    def get_index(self):
        """
        return the index object
        """
        return self.index

    def create_add_index_chroma(self, data, persist_directory=None):
        """
        Init the chromadb index and populate it with a set of documents
        """
        # init the ChromaDB client
        self.vdb_client = chromadb.PersistentClient(
            path=persist_directory,
             settings=Settings(
                allow_reset=True
            ))


        # make sure we're starting with a fresh db and collection
        self.vdb_client.reset()
        index = self.vdb_client.get_or_create_collection(name=self.index_name,
                                                         embedding_function=self.embeddings)

        # populate the collection
        self._add_docset_chroma(index, data)

        return index

    def add_to_index_chroma(self, data):
        """
        add document(s) to a chromadb index
        """
        # first, delete all existing docs from the same sources
        # as what we are adding, we don't want duplicates
        self._delete_docset_chroma(self.index, data)

        # now, add the new docs
        self._add_docset_chroma(self.index, data)

        # persist the db to disk
        # self.vdb_client.persist()

        return

    def add_to_index(self, data):
        """
        add document(s) to the agent's index
        """
        start_time = time.time()

        if self.index:
            if self.index_store == 'chroma':
                self.add_to_index_chroma(data)
                # force topk count to be at max count of index
                self.topk = min(self.topk, self.index.count())
            else:
                raise Exception(f"{self.index_store} does not support adding document")
        else:
            raise Exception("No available index, cannot add document")

        # log that the doc is added
        if self.index:
            duration = time.time() - start_time
            params = {
                "n_items": len(data),
            }
            event = self._log_event(agent_events._EVNT_INDEXADD, duration, params=params)

        return

    def create_add_index(self, data,
                         store='chroma',
                         persist_directory=None,
                         index_name=None):
        """
        create an index from a data source
        data: list of langchain Document() objects
        store: type of vectorstore to use (chroma, faiss, ...)
        """

        start_time = time.time()

        # note what store we are using and the index name
        self.index_store = store
        self.index_name = self.agent_name if not index_name else index_name

        # create the index
        if store == 'faiss':
            self.index = FAISS.from_documents(data, self.embeddings)
        elif store == 'chroma':
            self.index = self.create_add_index_chroma(data,
                                                      persist_directory=persist_directory)
            # force topk count to be at max count of index
            self.topk = min(self.topk, self.index.count())
        else:
            self.index = None
            self.index_store = None

        # log that the index is ready
        if self.index:
            duration = time.time() - start_time
            params = {
                "store": store,
                "persist_directory": persist_directory,
                "index_name": self.index_name,
                "n_items": len(data)
            }
            event = self._log_event(agent_events._EVNT_INDEXCREATE, duration, params=params)

        return

    def load_index(self, persist_directory, index_name, store='chroma'):
        """
        load an already persisted index from a directory
        persist_directory: location of persisted index
        store: type of vectorstore to use (chroma, ...)
                only supports chroma for now
        """
        start_time = time.time()

        # make note of the store type
        self.index_store = store

        # load the index
        if self.index_store == 'chroma':
            self.vdb_client = chromadb.PersistentClient(
                path=persist_directory,
                settings=Settings(
                    allow_reset=True
                ))
            self.index_name = index_name
            self.index = self.vdb_client.get_collection(name=self.index_name,
                                                          embedding_function=self.embeddings)
            # force topk count to be at max count of index
            self.topk = min(self.topk, self.index.count())
        else:
            self.index = None

        # log that the index is ready
        duration = time.time() - start_time
        params = {
            "store": self.index_store,
            "persist_directory": persist_directory,
        }
        event = self._log_event(agent_events._EVNT_INDEXLOAD, duration, params=params)

        return

    def get_index_stats(self):
        """
        return some stats about the agent's index
        """
        stats = None
        if self.index:
            try:
                stats = {
                    "name": self.index_name,
                    "store": self.index_store,
                    "n_items": self.index.count()
                }
            except:
                raise Exception("Index does not support stats")
        return stats

    def search_chromadb(self, query, filters=None, k=7, include_metadata=True):
        """
        run a search against the chromadb index for a list of queries
        """

        # set up the filters for metadata
        metadata_filters = None

        if filters == None:
            # no filters to apply
            metadata_filters = None
        else:
            # start a filter set
            valid_filters = False
            filter_set = []

            # check if we need to filter by page numbers
            if "pages" in filters:
                valid_filters = True
                pages = filters["pages"]
                page_offset = filters.get("page_offset", 0)
                page_tolerance = filters.get("page_tolerance", 1)
                # add page number filter
                filter_set.append({"page": {"$gte": pages[0]+page_offset-page_tolerance}})
                filter_set.append({"page": {"$lte": pages[-1]+page_offset+page_tolerance}})
                metadata_filters = {"$and": filter_set}

            if "_fieldtype" in filters:
                valid_filters = True
                fieldtypes = filters["_fieldtype"]
                if len(fieldtypes) > 1:
                    for f in fieldtypes:
                        filter_set.append({"_fieldtype": {"$eq": f}})
                    metadata_filters = {"$or": filter_set}
                else:
                    metadata_filters = {"_fieldtype": {"$eq": fieldtypes[0]}}

            ##
            ## other filters can go here
            ##

            ## REVISIT THIS
            # # collect all the filters into an AND conjuction
            # if valid_filters:
            #     if len(filter_set) > 1:
            #         metadata_filters = {"$and": filter_set}
            #     else:
            #         metadata_filters = filter_set

        # force k to be at most value of self.topk
        # self.topk will be max count of index
        k = min(k, self.topk)

        # perform query
        results = self.index.query(query_texts=[query],
                                    n_results=k,
                                    where=metadata_filters,
                                    where_document=None)

        # construct result docset
        docs = []
        for i in range(0, len(results['documents'][0])):
            page_content = results['documents'][0][i]
            metadata = results['metadatas'][0][i]
            distance = results['distances'][0][i]
            metadata['distance'] = distance
            doc = Document(page_content=page_content, metadata=metadata)
            docs.append(doc)

        return docs

    def _add_docset_chroma(self, index, data):
        """
        populate the index
        """
        if len(data) == 0:
            # cannot add anything
            return

        ids = [doc.metadata.get("id", self._create_id(f"{doc.page_content}{doc.metadata}")) for doc in data]
        docs = [doc.page_content for doc in data]
        metas = [doc.metadata for doc in data]

        index.add(
            documents=docs,
            metadatas=metas,
            ids=ids)

        return

    def _delete_docset_chroma(self, index, data):
        """
        delete from the index where source matches incoming data
        """
        ids = list(set([doc.metadata.get('id') for doc in data]))
        if len(ids) == 1:
            where_clause = {"id": ids[0]}
        else:
            where_clause = {
                "$or": [{"id": _id} for _id in ids]
            }
        index.delete(
            where=where_clause
        )
        return


    ## document search

    def get_similar_docs(self, query, filters=None, topk=7):
        """
        get top-K similar docs from an index given a query
        query: query string
        index: index object
        topk: number of top-K similar docs to matching query to return
        """

        if self.index:
            if self.index_store == 'faiss':
                docs = self.index.similarity_search(query,
                                                    k=topk,
                                                    include_metadata=True)
            elif self.index_store == 'chroma':
                docs = self.search_chromadb(query,
                                            filters=filters,
                                            k=topk,
                                            include_metadata=True)
            else:
                docs = None

        return docs



import json
import hashlib

__all__ = [
    'GoogleNewsSummarizer',
    'DatasetSummarizer'
]

class GoogleNewsSummarizer(object):
    """
    Class to handle summarization of Google News search results
    """

    def __init__(self):
        pass

    def create_id(self, string):
        """
        create a unique identifier from a string
        """
        return hashlib.md5(string.encode('utf-8')).hexdigest()

    def load_serp_results(self, file_path):
        """
        Take a json file path and load the results data
        """
        with open(file_path, "r") as fd:
            serp_results = json.load(fd)
        return serp_results

    def summarize_results(self, agent, serp_result, rulespec):
        """
        Construct the LLM prompt and call the agent to summarize
        """
        ## defaults
        default_ruleset = [
            "inlcude the source and URL for all news snippets you use to construct the summary using this format [source; url]"
        ]

        ## get the rules information
        # summary length
        summary_length = int(rulespec.get("summary_length", 100))

        # ruleset
        rule_set = rulespec.get("rule_set", default_ruleset)
        rule_set = [f"{ix}. {rule}" for ix, rule in enumerate(rule_set)]
        rule_set = "\n".join(rule_set)

        # cleaner method
        cleaner = rulespec.get("methods", {}).get("cleaner")
        crawler = rulespec.get("methods", {}).get("crawler")

        ## construct the prompt
        # get the news results
        for name in ['news', 'organic', 'video']:
            label = f"{name}_results"
            if label in serp_result:
                results = serp_result[label]
                break

        print("Results", len(results))
        # get the Google query
        g_query = serp_result["search_parameters"]["q"]
        # run the query cleaner method
        if callable(cleaner):
            g_query = cleaner(g_query)

        # get the snippets
        snippets = []
        count = 0
        for result in results:
            count += 1
            link = result["link"]
            title = result["title"]
            source = result.get("source", "Unknown")
            snippet = result.get("original_snippet")
            if snippet == None:
                snippet = result.get("snippet", "")
            content = result.get("snippet", "")
            # run the crawler method
            if callable(crawler):
                snippet = crawler(link)

            entry = f"""
                [Result Number: {count}]
                Title: {title}
                Snippet: {snippet}
                Source: {source}
                URL: {link}
            """
            # take care of the case where snippet has been overwritten with original_snippet
            if "original_snippet" in result:
                entry = f"""{entry}
                Content: {content}
                """
            snippets.append(entry)
        snippets = "\n\n".join(snippets)

        prompt = {
            "persona": """You are a highly advanced AI capable of summarizing Google News and other search result snippets.""",
            "prompt": f"""
                ## Input
                The following are a set of {len(results)} news results snippets and crawled page content from the results of a Google query:
                {snippets}

                The query terms issued to Google were: {g_query}

                ## Your task is to summarize the news results snippets for a reader in {summary_length} words or less. Pay special attention to the following rules:
                {rule_set}

                ## Your summary:"""
        }

        ## run prompt
        # call the LLM with the prompt
        response = agent.prompt(prompt)

        return response

    def process_results(self, agent, serp_results, rulespec):
        """
        Process all the search results for summarization
        """
        for serp_result in serp_results:

            if not isinstance(serp_result, dict):
                continue

            checks = [f"{name}_results" in serp_result \
                          for name in ['news', 'organic', "video"]]
            if not any(checks):
                continue

            response = self.summarize_results(agent, serp_result, rulespec)
            summary = response["answer"]

            try:
                serp_result["summary"] = json.loads(summary)
            except:
                serp_result["summary"] = summary

        return serp_results

    def dedup_results(self, agent, serp_results):
        """
        De-duplicate the news results
        """

        # create an id for each summary text
        snippets = []
        for r in serp_results:
            summary = r.get('summary', {})
            texts = summary.get('summary', [])
            if not isinstance(texts, list):
                continue
            if len(texts) > 0:
                for text in texts:
                    if not isinstance(text, dict):
                        continue
                    t = text.get('text', "")
                    text['id'] = self.create_id(t)
                    snippets.append({
                        "id": text['id'],
                        "text": t,
                    })

        entries = []
        for s in snippets:
            entry = f"""
                ID: {s['id']}
                Snippet: {s['text']}
            """
            entries.append(entry)
        entries = "\n\n".join(entries)

        prompt = {
            "persona": """You are a highly advanced AI capable of deduplicating snippets of text.""",
            "prompt": f"""
                ## Input
                The following are a set of {len(entries)} text snippets:
                {entries}

                ## Your task is to deduplicate the set of text snippets.
                If there are multiple text snippets with the same meaning, inlcude only the first one in your response.
                If a text snippet contains any variation of the phrase "Nothing Relevant", do not include it in your response.

                Format your response as a list of JSON objects as """ + "{id: ID, text: snippet}" + """
                DO NOT include code tags.

                ## Your response:"""
        }


        ## run prompt
        # call the LLM with the prompt
        response = agent.prompt(prompt)
        try:
            results = json.loads(response['answer'])
        except:
            print("Received unexpected response")
            print(response)
            raise

        valid_stories = {result.get("id"): result.get("text") for result in results}

        # go through the summary texts and drop duplicates
        # duplicate texts will not be found in valid_stories
        for r in serp_results:
            summary = r.get('summary', {})
            texts = summary['summary']
            if not isinstance(texts, list):
                continue
            texts = [t for t in texts if (isinstance(t, dict) and ('id' in t))]
            if len(texts) > 0:
                texts = [text for text in texts if text["id"] in valid_stories]
                summary['summary'] = texts

        return serp_results

    def store_results(self, serp_results, out_file_path):
        """
        Store the summary results
        """
        with open(out_file_path, "w") as fd:
            json.dump(serp_results, fd, indent=4)

class DatasetSummarizer(object):
    """
    Class to handle summarization of dataset
    """
    def load_dataset(self, inputfile):
        """
        Load the dataset. For now it is not doing anything interesting but
        may be it will do something useful in future
        """

        if not inputfile.lower().endswith(".csv"):
            raise Exception("Only csv files supported")

        return open(inputfile).read()

    def process_results(self, agent, data, rulespec):

        ## defaults
        default_ruleset = [
            "Analyze the dataset and summarize observations"
        ]

        # summary length
        summary_length = int(rulespec.get("summary_length", 100))
        description = rulespec.get("description", "")

        # ruleset
        rule_set = rulespec.get("rule_set", default_ruleset)
        rule_set = [f"{ix}. {rule}" for ix, rule in enumerate(rule_set)]
        rule_set = "\n".join(rule_set)

        prompt = {
            "persona": """You are a highly advanced AI capable of analyzing datasets and providing accurate and relevant answers""",
            "prompt": f"""
                {description}

                ## Input
                The following is a set of records in CSV format:
                {data}

                ## Your task is to summarize the records for a reader in {summary_length} words or less. Pay special attention to the following rules:
                {rule_set}

                ## Your summary:"""
        }

        ## run prompt
        # call the LLM with the prompt
        response = agent.prompt(prompt)

        return response

    def store_results(self, result, outputpath):
        """
        Store the summary results
        """
        with open(outputpath, "w") as fd:
            json.dump(result, fd, indent=4)



import os
import sys
import json
from fastapi.concurrency import run_in_threadpool

from .basegpt import BaseGPT
from .lib import *
from ..agents import LLMDatabankQuerier

logger = get_logger()

class DatabankQueryGPT(BaseGPT):

    def __init__(self):
        super().__init__()
        self.name = "DatabankQueryGPT"
        self.agentname = os.environ.get('AGENTNAME', 'databankindex')
        self.platform = "azure" if 'AZURE_OPENAI_ENDPOINT' in os.environ else "openai"

        self.model = os.environ.get("MODELNAME", 'gpt-4o-mini')
        self.namespace = self.agentname
        self.index_name = "databank_gpt_index"
        self.router.add_api_route("/load", self.load_data, methods=["POST"])

        self.persist_directory = os.path.join(self.run_dir,
                                              self.agentname,
                                              "index")
        
        stats = get_stats()
        stats.update({
            "model": self.model,
            "platform": self.platform, 
            "agentname": self.agentname,
        })
        
    async def get_llm_agent(self):
        agent = LLMDatabankQuerier(name=self.agentname,
                                   model=self.model,
                                   platform=self.platform)

        logger.debug(f"Built agent for {self.agentname} w/ model {self.model} and platform {self.platform}")

        return {
            "agent": agent
        }

    async def qna_run(self, request_id):
        """
        Execute query against the agend
        :param request_id: Request ID of the request
        :return:
        """
        cache = get_cache()

        if request_id not in cache:
            logger.error(f"Failure",
                         extra={
                             'request_id': "invalid",
                             "source": "service",
                         })
            cache[request_id] = {
                'status': "failure",
                "message": f"Invalid request id"
            }
            return

        # First get the params
        value = cache[request_id]

        try:

            params  = value['params']
            user    = params['user']
            dataset = params['dataset']
            query   = params['query']
            extra   = params['extra']

            stats['query_count'] += 1

            label = f"{user}_{dataset}"

            stats['datasets'][label] = {
                'loaded': datetime.now().replace(microsecond=0).isoformat(),
                'username': user,
                'agent_created': True,
                'agent_status': "Created",
                "query_count": 0,
                "query_success": 0,
                "query_failure": 0,
            }
            
            # First get the agent...
            logger.debug(f"Running Query",
                         extra={
                             "source": "service",
                             "user": user,
                             "dataset": dataset,
                             "request_id": request_id,
                             'data': json.dumps(value, indent=4, cls=SafeEncoder)
                         })

            index_name = self.index_name
            index_path = os.path.join(self.persist_directory, "index")
            if ((self.agent.index is None) and (os.path.exists(index_path))):
                self.agent.load_index(store="chroma",
                                      persist_directory=self.persist_directory,
                                      index_name=index_name)
                logger.debug(f"Loaded index",
                             extra={
                                 "data": json.dumps({
                                     'source': 'service',
                                     'user': 'all',
                                     'dataset': dataset,
                                     'stats': self.agent.get_index_stats()
                                 }, indent=4, cls=SafeEncoder)
                             })            

            result = await run_in_threadpool(lambda: self.agent.query(query,
                                                                      filters=extra.get('filters',{}),
                                                                      topk=extra.get('topk', 3)))

            if result.get("success"):
                key = result.get("key")
                logger.debug(f"Query: {query}\nKEY: {key}\n")
                json_result = json.loads(json.dumps(result, indent=4, cls=SafeEncoder))

                stats['query_success'] += 1
                stats['datasets'][label]['query_count'] += 1
                stats['datasets'][label]['query_success'] += 1
                query_update_result(request_id, {
                    "status": "success",
                    "result": json_result,
                })
            else:
                suggestions = result.get("suggestions")
                logger.debug(f"Query: {query}\nSuggestions: {suggestions}\n")
                json_result = json.loads(json.dumps(result, indent=4, cls=SafeEncoder))
                stats['datasets'][label]['query_count'] += 1
                stats['datasets'][label]['query_failure'] += 1
                query_update_result(request_id, {
                    "status": "success",
                    "result": json_result,
                })

        except Exception as e:
            stats['query_failure'] += 1
            stats['datasets'][label]['query_count'] += 1
            stats['datasets'][label]['query_failure'] += 1
            logger.exception(f"Failed to run query",
                             extra={
                                 "source": "service",
                                 'request_id': request_id,
                                 "user": params.get('user', "unknown"),
                                 "dataset": params.get('dataset', "unknown"),
                             })
            query_update_result(request_id, {
                "status": "failure",
                "answer": f"Unable to construct the answer. Could be an internal agent error. See the agent log",
                "result": {
                    "text": str(e)
                }
            })

    async def load_data(self,
                        data: list[dict]= [],
                        params: dict = {}):
        """
        Load JSON-like structure

        :param data: List of dictionaries
        :param params: Indexing params
        """
        logger.debug(f"Loading data records {len(data)} {params}")

        try:
            loadfunc = lambda: self.agent.load_data(content=data,
                                                    source='json',
                                                params=params)
            data = await run_in_threadpool(loadfunc)
        except Exception as e:
            logger.exception("Failed to turn dicts into docs")
            raise HTTPException(
                status_code=500,
                detail=f'Failed to load data: {e}'
            )

        try:
            print("Loading index", self.index_name, self.persist_directory)
            ## index the data
            # create the index and add data to it
            # this call will flush an existing index if it already exists
            # can be called multiple times
            indexfunc = lambda: self.agent.create_add_index(data=data,
                                                            store="chroma",
                                                            persist_directory=self.persist_directory,
                                                            index_name=self.index_name)
            data = await run_in_threadpool(indexfunc)
        except Exception as e:
            raise HTTPException(
                status_code=500,
                detail=f'Failed to index loaded data: {e}'
            )            

        return {
            "status": "success"
        }
    
app = FastAPI()

##############################################
# Common
##############################################
add_show_cache(app)
add_show_health(app)
databankGpt = DatabankQueryGPT()
app.include_router(databankGpt.router)

# For IDE Debugging
if __name__ == "__main__":
    uvicorn.run(app, host="0.0.0.0", port=10892)



import os
import re
import requests
import logging
from urllib.parse import urlencode, unquote_plus

logger = logging.getLogger("app")

__all__ = ['AzureDynamicsClient']

#
# AzureDynamicsClient to invoke Dynamics365 web apis.
#
class AzureDynamicsClient:
    api_path = "api/data/v9.2"

    def __init__(self, domain, tenant_id, client_id, client_secret, login_url, grant_type, access_token=None):
        self.domain = domain.strip("/")
        self.tenant_id = tenant_id
        self.client_id = client_id
        self.client_secret = client_secret
        self.login_url = login_url
        self.grant_type = grant_type

        self.headers = {
            "Accept": "application/json, */*",
            "content-type": "application/json; charset=utf-8",
        }
        if access_token is not None:
            self.set_access_token(access_token)

    #
    # Sets the access token for use in this library in the header information to make request calls.
    # This is  obtained after successful login.
    #
    def set_access_token(self, token):
        assert token is not None, "The token cannot be None."
        self.access_token = token
        self.headers["Authorization"] = "Bearer " + self.access_token

    #
    # Does the login to get the access token and sets it for use in the library.
    #
    def do_login(self):
        # build the authorization token request
        tokenpost = {
            'tenant_id' : self.tenant_id,
            'client_id': self.client_id,
            'client_secret': self.client_secret,
            'grant_type': self.grant_type,
            'resource': self.domain
        }
        # make the login request
        try:
            response = requests.post(self.login_url, data=tokenpost)
        except Exception as e:
            msg = "Login request received exception: {0}".format(e)
            logger.error(msg)
            raise Exception(msg)

        # check the response code
        if response.status_code != 200:
            msg = "Login request could not get access token. STATUS_CODE: {0}, URL: {1}, MESSAGE: {2}".format(
                response.status_code, response.url, response.text)
            logger.error(msg)
            raise Exception(msg)
        else:
            pacctok = response.json().get('access_token')
            self.set_access_token(pacctok)
            logger.info("Login request obtained access token.")
            return {
                'status': 'success',
                'access_token': pacctok,
            }

    #
    # Method to make the request, receives the different methods (post, delete, patch, get) that the api allows,
    # see the documentation to check how to use the filters:
    # https://msdn.microsoft.com/en-us/library/gg309461(v=crm.7).aspx
    #
    def make_request(
        self,
        method,
        endpoint,
        expand=None,
        filter=None,
        orderby=None,
        select=None,
        skip=None,
        top=None,
        data=None,
        json=None,
        **kwargs,
    ):
        extra = {}
        if expand is not None and isinstance(expand, str):
            extra["$expand"] = str(expand)
        if filter is not None and isinstance(filter, str):
            extra["$filter"] = filter
        if orderby is not None and isinstance(orderby, str):
            extra["$orderby"] = orderby
        if select is not None and isinstance(select, str):
            extra["$select"] = select
        if skip is not None and isinstance(skip, str):
            extra["$skip"] = skip
        if top is not None and isinstance(top, str):
            extra["$top"] = str(top)

        assert self.domain is not None, "'domain' is required"
        assert self.access_token is not None, "You must provide a 'token' to make requests"
        url = f"{self.domain}/{self.api_path}/{endpoint}?" + urlencode(extra)
        msg = "Make request method: {0} url: {1} json: {2}".format(method, unquote_plus(url), json)
        logger.info(msg)
        try:
            if method == "get":
                response = requests.request(method, url, headers=self.headers, params=kwargs)
            else:
                response = requests.request(method, url, headers=self.headers, data=data, json=json)
        except Exception as e:
            msg = "Make request received exception: {0}".format(e)
            logger.error(msg)
            raise Exception(msg)

        return self.parse_response(response)

    #
    # Wrapper to make get request.
    #
    def _get(self, endpoint, data=None, **kwargs):
        return self.make_request("get", endpoint, data=data, **kwargs)

    #
    # Wrapper to make post request.
    #
    def _post(self, endpoint, data=None, json=None, **kwargs):
        return self.make_request("post", endpoint, data=data, json=json, **kwargs)

    #
    # Wrapper to make delete request.
    #
    def _delete(self, endpoint, **kwargs):
        return self.make_request("delete", endpoint, **kwargs)

    #
    # Wrapper to make patch request.
    #
    def _patch(self, endpoint, data=None, json=None, **kwargs):
        return self.make_request("patch", endpoint, data=data, json=json, **kwargs)

    #
    # Parse the response and provide the json data if successful or raise exception
    #
    def parse_response(self, response):
        if response.status_code == 200:
            return response.json()
        elif response.status_code == 204 or response.status_code == 201:
            if 'OData-EntityId' in response.headers:
                entity_id = response.headers['OData-EntityId']
                if entity_id[-38:-37] == '(' and entity_id[-1:] == ')':  # Check container
                    guid = entity_id[-37:-1]
                    guid_pattern = re.compile(r'^[\da-f]{8}-([\da-f]{4}-){3}[\da-f]{12}$', re.IGNORECASE)
                    if guid_pattern.match(guid):
                        return guid
                    else:
                        return True  # Not all calls return a guid
            else:
                return True
        else:
            msg = "Request received error. STATUS_CODE: {0}, URL: {1}, MESSAGE: {2}".format(
                response.status_code, response.url, response.text)
            logger.error(msg)
            raise Exception(msg)

    #
    # Generic methods to operate on the entity type data.
    #
    def get_data(self, type=None, **kwargs):
        if type is not None:
            return self._get(type, **kwargs)
        msg = "A type is necessary. Example: contacts, leads, accounts, etc... check the library"
        logger.error(msg)
        raise Exception(msg)

    def create_data(self, type=None, params=None):
        if type is not None:
            if params is not None and isinstance(params, dict):
                return self._post(type, json=params)
            else:
                msg = "params needs to be a dict"
                logger.error(msg)
                raise Exception(msg)
        msg = "A type is necessary. Example: contacts, leads, accounts, etc... check the library"
        logger.error(msg)
        raise Exception(msg)

    def update_data(self, type=None, id=None, params=None):
        if type is not None and id is not None:
            url = "{0}({1})".format(type, id)
            if params is not None and isinstance(params, dict):
                return self._patch(url, json=params)
            else:
                msg = "params needs to be a dict"
                logger.error(msg)
                raise Exception(msg)
        msg = "A type is necessary. Example: contacts, leads, accounts, etc... check the library"
        logger.error(msg)
        raise Exception(msg)

    def delete_data(self, type=None, id=None):
        if type is not None and id is not None:
            return self._delete("{0}({1})".format(type, id))
        msg = "A type is necessary. Example: contacts, leads, accounts, etc... check the library"
        logger.error(msg)
        raise Exception(msg)

    #
    # contact section, see the documentation
    # https://docs.microsoft.com/es-es/dynamics365/customer-engagement/web-api/contact?view=dynamics-ce-odata-9
    #
    def get_contacts(self, **kwargs):
        return self._get("contacts", **kwargs)

    def create_contact(self, params):
        if params is not None and isinstance(params, dict):
            return self._post("contacts", json=params)
        else:
            msg = "params needs to be a dict"
            logger.error(msg)
            raise Exception(msg)

    def delete_contact(self, id):
        if id != "":
            return self._delete("contacts({0})".format(id))
        msg = "ID is required to delete a contact"
        logger.error(msg)
        raise Exception(msg)

    def update_contact(self, id, params):
        if id != "":
            url = "contacts({0})".format(id)
            if params is not None and isinstance(params, dict):
                return self._patch(url, json=params)
            else:
                msg = "params needs to be a dict"
                logger.error(msg)
                raise Exception(msg)
        msg = "ID is required to update a contact"
        logger.error(msg)
        raise Exception(msg)

    #
    # account section, see the documentation
    # https://docs.microsoft.com/es-es/dynamics365/customer-engagement/web-api/account?view=dynamics-ce-odata-9
    #
    def get_accounts(self, **kwargs):
        return self._get("accounts", **kwargs)

    def create_account(self, params):
        if params is not None and isinstance(params, dict):
            return self._post("accounts", json=params)
        else:
            msg = "params needs to be a dict"
            logger.error(msg)
            raise Exception(msg)

    def delete_account(self, id):
        if id != "":
            return self._delete("accounts({0})".format(id))
        msg = "ID is required to delete an account"
        logger.error(msg)
        raise Exception(msg)

    def update_account(self, id, params):
        if id != "":
            url = "accounts({0})".format(id)
            if params is not None and isinstance(params, dict):
                return self._patch(url, json=params)
            else:
                msg = "params needs to be a dict"
                logger.error(msg)
                raise Exception(msg)
        msg = "ID is required to update an account"
        logger.error(msg)
        raise Exception(msg)

    #
    # opportunity section, see the documentation
    # https://docs.microsoft.com/es-es/dynamics365/customer-engagement/web-api/opportunity?view=dynamics-ce-odata-9
    #
    def get_opportunities(self, **kwargs):
        return self._get("opportunities", **kwargs)

    def create_opportunity(self, params):
        if params is not None and isinstance(params, dict):
            return self._post("opportunities", json=params)
        else:
            msg = "params needs to be a dict"
            logger.error(msg)
            raise Exception(msg)

    def delete_opportunity(self, id):
        if id != "":
            return self._delete("opportunities({0})".format(id))
        msg = "ID is required to delete an opportunity"
        logger.error(msg)
        raise Exception(msg)

    def update_opportunity(self, id, params):
        if id != "":
            url = "opportunities({0})".format(id)
            if params is not None and isinstance(params, dict):
                return self._patch(url, json=params)
            else:
                msg = "params needs to be a dict"
                logger.error(msg)
                raise Exception(msg)
        msg = "ID is required to update an opportunity"
        logger.error(msg)
        raise Exception(msg)

    #
    # leads section, see the documentation
    # https://docs.microsoft.com/es-es/dynamics365/customer-engagement/web-api/lead?view=dynamics-ce-odata-9
    #
    def get_leads(self, **kwargs):
        return self._get("leads", **kwargs)

    def create_lead(self, params):
        if params is not None and isinstance(params, dict):
            return self._post("leads", json=params)
        else:
            msg = "params needs to be a dict"
            logger.error(msg)
            raise Exception(msg)

    def update_lead(self, id, params):
        if id != "":
            url = "leads({0})".format(id)
            if params is not None and isinstance(params, dict):
                return self._patch(url, json=params)
            else:
                msg = "params needs to be a dict"
                logger.error(msg)
                raise Exception(msg)
        msg = "ID is required to update a lead"
        logger.error(msg)
        raise Exception(msg)

    def delete_lead(self, id):
        if id != "":
            return self._delete("leads({0})".format(id))
        msg = "ID is required to delete a lead"
        logger.error(msg)
        raise Exception(msg)

    #
    # campaign section, see the documentation
    # https://docs.microsoft.com/es-es/dynamics365/customer-engagement/web-api/campaign?view=dynamics-ce-odata-9
    #
    def get_campaigns(self, **kwargs):
        return self._get("campaigns", **kwargs)

    def create_campaign(self, params):
        if params is not None and isinstance(params, dict):
            return self._post("campaigns", json=params)
        else:
            msg = "params needs to be a dict"
            logger.error(msg)
            raise Exception(msg)

    def update_campaign(self, id, params):
        if id != "":
            url = "campaigns({0})".format(id)
            if params is not None and isinstance(params, dict):
                return self._patch(url, json=params)
            else:
                msg = "params needs to be a dict"
                logger.error(msg)
                raise Exception(msg)
        msg = "ID is required to update a campaign"
        logger.error(msg)
        raise Exception(msg)

    def delete_campaign(self, id):
        if id != "":
            return self._delete("campaigns({0})".format(id))
        msg = "ID is required to delete a campaign"
        logger.error(msg)
        raise Exception(msg)

if __name__ == "__main__":

    # NOTE: Ensure that logs/ directory is present, else will get an error while running.
    logging.basicConfig(filename='logs/azuredynamicsclient.log',
                        level=logging.INFO,
                        format='%(asctime)s %(levelname)s %(module)s.%(funcName)s:%(lineno)d %(message)s')

    # Read the env variables necessary to retrieve the oauth token.
    # Refer to readme.md file for details.
    tenant_id       = os.environ["SCR_CRM_TENANT_ID"]
    client_id       = os.environ["SCR_CRM_CLIENT_ID"]
    client_secret   = os.environ["SCR_CRM_CLIENT_SECRET"]
    domain          = os.environ["SCR_CRM_DOMAIN"]
    login_url       = os.environ["SCR_CRM_LOGIN_URL"]
    grant_type      = "client_credentials"

    # Create the crm client. access_token is setup in the do_login step.
    crmclient = AzureDynamicsClient(domain=domain, tenant_id=tenant_id,
                    client_id=client_id, client_secret=client_secret,
                    login_url=login_url, grant_type=grant_type, access_token=None)

    # Do the login to set the access token.
    resp = crmclient.do_login()
    if resp['status'] == "success":
        ############## contacts operations ######################
        # Perform get contacts
        print("Getting contacts...[contactid] - [fullname]")
        crmresults = crmclient.get_contacts()
        # Loop through it
        for x in crmresults['value']:
            print (x['contactid'] + ' - ' + x['fullname'])

        # Perform a create contact
        print("\nCreating contact...[firstname:test fname]")
        contact_data = {"firstname": "test fname",
                        "lastname": "test lname",
                        "middlename": "test mname",
                        "emailaddress1": "testcontact@test.com",
                        "telephone1": "654-321-0123"}
        # Store created contactid for further uses
        pcontactid = crmclient.create_contact(contact_data)
        print("Result created contactid:", pcontactid)

        # Perform a get contact with select fields
        print("\nGetting contacts with select fields...[firstname] - [telephone1]")
        crmresults = crmclient.get_contacts(select="firstname,telephone1")
        # Loop through it
        for x in crmresults['value']:
            print (x['contactid'] + ' - ' + x['firstname'] + ' - ' + x['telephone1'])

        # Perform a get contact with filter
        print("\nGetting filtered contact...[contactid], select fields [firstname] - [telephone1]")
        filterstr = f"contactid eq {pcontactid}"
        crmresults = crmclient.get_contacts(filter=filterstr,
                                            select="firstname,telephone1")
        # Loop through it
        for x in crmresults['value']:
            print (x['contactid'] + ' - ' + x['firstname'] + ' - ' + x['telephone1'])

        # Perform an update contact
        print("\nUpdating contact...[contactid] - [telephone1 654-777-7777]")
        updcontact_data = {"telephone1": "654-777-7777"}
        crmresults = crmclient.update_contact(id=pcontactid, params=updcontact_data)
        print("Result:", crmresults)

        # Perform a get contact with filter
        print("\nGetting filtered contact...[contactid], select fields [firstname] - [telephone1]")
        filterstr = f"contactid eq {pcontactid}"
        crmresults = crmclient.get_contacts(filter=filterstr,
                                            select="firstname,telephone1")
        # Loop through it
        for x in crmresults['value']:
            print (x['contactid'] + ' - ' + x['firstname'] + ' - ' + x['telephone1'])

        # Perform a delete contact
        print("\nDeleting contact...[contactid]")
        crmresults = crmclient.delete_contact(id=pcontactid)
        print("Result:", crmresults)

        ############## leads operations ######################
        # Perform a sample get leads
        print("\nGetting leads...[fullname] - [leadid] - [subject]")
        crmresults = crmclient.get_leads()
        # Loop through it
        for x in crmresults['value']:
            print (x['fullname'] + ' - ' + x['leadid'] + ' - ' + x['subject'])

    else:
        print("Unable to login.")



import os
import json
import pandas as pd

from langchain_core.prompts import PromptTemplate
from langchain_openai import ChatOpenAI
from langchain_core.messages import HumanMessage, SystemMessage
from langchain_community.callbacks.manager import get_openai_callback

import logging
from logging.config import dictConfig
import time
import hashlib
import unittest

from . import agent_events
from llmsdk.lib import defaults
from llmsdk.lib.enablers import AgentEnablersMixin
try:
    from llmsdk.services.log import *
    logging.config.dictConfig(log_config)
except:
    logging.basicConfig()

__all__ = ['SADLClassifier']

class SADLClassifier(AgentEnablersMixin):
    """
    Class to take in a dataframe and create a data dictionary for it
    """

    def __init__(self,
                 cred={},
                 platform=defaults.LLM_PLATFORM,
                 model=defaults.LLM_MODEL,
                 entities_filepath=None,
                 max_tokens=1024,
                 temperature=0):
        """
        Class to classify data columns using SADL

        Parameters:
        - entities_filepath (str): The file path to the entities mapping file in JSON format. Default is None.
        - temperature (int): Control the temperature of the inference/API
        - max_tokens (int): Control the max_tokens in the inference/API

        Output:
        Use classify_columns method
        """

        start_time = time.time()

        # init
        self.context = None
        self.agent_name = "sadl"
        self.max_tokens = max_tokens
        self.temperature = temperature
        self.data = None
        self.samplesize = 5
        self.platform = platform
        self.model = model


        # init the base entity map
        self.entity_map = self._get_base_entity_mapping()

        # init the base industry list
        self.industries = self._get_base_industries()

        # load additional entities if needed
        if entities_filepath:
            self.entity_map = self.load_entities(entities_filepath)

        # init the llm and embeddings objects
        self.llm, self.embeddings = self._get_llm_objs(platform=self.platform,
                                                          model=self.model,
                                                          embedding_model=self.embedding_model,
                                                          cred=self.cred)

        self.logger = logging.getLogger("app")

        self.agent_id = self._create_id(f"{self.agent_name}_{start_time}")

        # set metadata
        self.metadata = {
            "agent": {
                "name": self.agent_name,
                "id": self.agent_id,
                "platform": self.platform,
                "model": self.model
            },
            "events": []
        }

        # log that the agent is ready
        duration = time.time() - start_time
        event = self._log_event(agent_events._EVNT_READY, duration)


    def _get_base_industries(self):
        """
        give the base industries
        """
        industries = ["Business", "Technology",
                           "Finance", "Healthcare",
                           "Manufacturing", "Retail",
                           "Government", "eCommerce",
                           "Education", "Logistics"]

        return industries

    def _get_base_entity_mapping(self):
        """
        give the default base mapping of entities
        """
        mapping = {
            "person": ["id", "name", "gender", "occupation", "nationality", "age", "marital status", "education level", "salutation", "job title", "relationship", "other"],
            "location": ["id", "city", "state", "province", "country", "address", "postal code", "latitude and longitude", "landmark", "neighborhood", "region", "time zone", "other"],
            "organization": ["id", "type", "industry", "other"],
            "event": ["id", "type", "theme", "other"],
            "product": ["id", "category", "brand", "model", "feature", "other"],
            "service": ["id", "category", "cost", "feature", "other"],
            "date/time": ["id", "date", "day", "month", "year", "timestamp", "time zone", "duration", "other"],
            "identifier": ["id"],
            "metric": ["revenue", "size", "volume", "sales", "measurement", "price", "other"],
            "contact": ["website", "email", "phone number", "social link"]
        }
        return mapping


    ## interfaces

    def load_entities(self, entities):
        """
        load any additional entity mappings and add them to the
        base entity map
        entities: path to entity mappings
        """

        # get the base entity map
        entity_map = self.entity_map

        # check path
        if os.path.exists(entities) and entities.endswith(".json"):
            with open(entities, "r") as fd:
                addl_emap = json.load(fd)
            if isinstance(addl_emap, dict):
                for  e, se in addl_emap.items():
                    # lowercase everything
                    e = e.lower()
                    # we need lists
                    if isinstance(se, str):
                        se = [se.lower()]
                    # lowercase everything
                    if isinstance(se, list):
                        se = [i.lower() for i in se]
                    else:
                        # we cannot add this item, move on
                        continue
                    # we can now add this mapping to the base map
                    entity_map[e].extend(se)
                    # make sure we remove duplicates
                    entity_map[e] = list(set(entity_map[e]))
            # done, we have the updated map
            self.entity_map = entity_map
        else:
            pass

        return self.entity_map

    def generate_prompt_params(self, context=""):
        def construct_promptpartial_from_map(entity_map):
            """
            Take an entity map and construct the entity mapping partial
            that goes into the prompt for the LLM
            """
            # first the entities
            entities = ", ".join(list(entity_map.keys()))
            partial_e = f"Assume the entities available for classification are as follows: \n{entities}"

            # then for the sub-entities
            partial_se = ""
            for e, se in entity_map.items():
                p_se = f"Assume the following are the sub entities for {e}: \n{', '.join(se)}"
                partial_se = f"{partial_se}\n\n{p_se}"

            partial = f"{partial_e}{partial_se}"

            return partial

        def construct_promptpartial_outputformat():
            """
            Construct the output format structure
            for the response from the LLM
            """
            column_label = "column_name"
            label_fields = ['datatype', 'entity', 'sub_entity', 'pii', 'description']

            p = ", ".join([f'"{f}": {f}' for f in label_fields])
            op_format = "{" + column_label + ": {" + p + "}}"

            return op_format


        # construct the prompt partial for the entity mappings
        map_partial = construct_promptpartial_from_map(self.entity_map)

        # get the input column name labels
        labels = ", ".join(list(self.df.columns))

        # output format
        output_format = construct_promptpartial_outputformat()

        resp = {"context": context,
             "map_partial": map_partial,
             "labels": labels,
             "output_format": output_format
            }

        return resp

    def generate_prompt_columns_string(self):
        """
        generate prompt in string format
        """
        prompt = self.generate_prompt_columns()
        params = self.generate_prompt_params()
        prompt_string = prompt.invoke(params)
        return prompt_string

    def generate_prompt_columns(self):
        """
        generate a prompt for labelling a dataframe given some context
        """
        prompt = ChatPromptTemplate.from_messages(
            [
                (
                    "system",
                    """You are a data entry operator.
                    Your task is to construct a data dictionary for a set of database column names given to you.
                    The database is from a company in {context} industry.""",
                ),
                ("human",

            """{map_partial}

            Classify the following column names into entity and sub_entity using the above mentioned details.
            Also provide the following for each column:
            - datatype
            - true/false indicating whether the column could contain personally identifiable information (PII)
            - description

            Format your output as a nested json dictionary as follows:
            {output_format}

            Here are the input column names:
            {labels}"""),
            ]
        )

        return prompt

    def generate_prompt_industry_params(self, df):
        # get the input labels
        labels = ", ".join(list(df.columns))
        params = {
            "labels": labels,
            "industries": self.industries
        }
        return params

    def generate_prompt_industry(self):
        """
        generate a prompt for identifying the industry of a dataframe given some context
        """
        prompt = ChatPromptTemplate.from_messages(
            [
                (
                    "system",
                    """You are a data entry operator.
                    Assume you have a list of industries:
                    {industries}""",
                ),
                (
                    "human",

                    """Your task is to identify what industry a dataset belongs to given the columns in the dataset.
                    Respond with EXACTLY one option from the list of industries.

                    Here are the columns in the dataset:
                    {labels}"""),
            ]
        )

        return prompt

    def load_data(self, content, source="df"):
        """
        - source (str): The type of input data. Can be "df" for a DataFrame input or "csv" for a CSV file input. Default is "df".
        - content: data file path (if source=='str') or DataFrame (if source=='df').
        """
        if source == "csv":
            data = pd.read_csv(content,
                                    on_bad_lines='skip',
                                    encoding='unicode_escape')
        elif source == "df" and not content.empty:
            data = content
        else:
            return False

        self.data = {}
        for col in data.columns:
            self.data[col] = list(data[col].astype(str).head(self.samplesize).values)

        return self.data

    def classify_columns(self, context=""):
        """
        Classify the columns.

        Parameters:
        - context (str, optional): The context for classification. If not provided, the default context of the class instance is used.

        Returns:
        - success (bool): Indicates whether the classification was successful.
        - result_type (str): The type of the classification result. Can be 'json' for JSON format or 'str' for string format.
        - result (str or dict): The classification result. If 'result_type' is 'json', it's a dictionary; otherwise, it's a string.
        """

        # set start_time
        start_time = time.time()

        # set the context for this labelling attempt
        context = self.context if context=="" else context

        # get the prompt for the LLM
        prompt = self.generate_prompt_columns()

        params = self.generate_prompt_params(context)

        chain = prompt | self.llm

        try:
            # chain and prompt
            success = True
            resp = chain.invoke(params)
            response = resp.content

            try:
                result = json.loads(response.lower().strip())
                result_type = 'json'
            except:
                result = response.strip()
                result_type = 'str'
        except:
            success = False
            result = ""
            result_type = ""

        # get prompt
        prompt_string = self.generate_prompt_columns_string()

        params = {
            "query": str(prompt_string),
            "mode": self.data_mode,
            "success": success,
            "result_type": result_type,
            "result": result
        }

        # logging the result
        duration = time.time() - start_time

        event = self._log_event(agent_events._EVNT_QUERY, duration, params=params)

        # process the response
        return success, result_type, result

    def classify_industry(self):
        """
        figure out what industry the dataframe is from
        """

        # get the prompt for the LLM
        prompt = self.generate_prompt_industry()

        params = self.generate_prompt_industry_params(self.df)

        chain = prompt | self.llm

        # chain
        resp = chain.invoke(params)
        response = resp.content

        result = response.lower().strip()

        # process the response
        return result

    def map_to_targets(self, data, targets, use_content=False):
        """
        map input column names to a defined set of target classes
        - data: input data to map, must come from the self.load_data(...) method
        - targets: list of target classes
        - use_content: if True, send a sample of data content values to LLM to do the mapping
            set this to True using caution, or data leakage is possible
        """

        def construct_prompt(data, targets, use_content):
            if use_content == False:
                task_subprompt = """You will be given a list of input columns as well as a list of target classes.
                Your task is to map each column in the input dataframe to one entry in the list of target classes.
                """
            if use_content == True:
                task_subprompt = """You will be given a column name, some sample values from that column, as well as a list of target classes.
                Your task is to map the column name to one entry in the list of target classes. Use the sample values to guide your decision.
                """


            # construct the prompt template
            # this is the system message part
            sys_msg = f"""You are a highly advanced, AI-enabled, data mapping tool.
            {task_subprompt}
            Format your output as a json dictionary as follows:
            {{"input": "target"}}
            """

            # this is the human message part
            if use_content == False:
                human_subprompt = f"""Here are the input columns names
                ------ BEGIN COLUMN NAMES ------
                {data}
                ------- END COLUMN NAMES -------
                """
            if use_content == True:
                human_subprompt = f"""Here is the column name and values
                ------ BEGIN COLUMN AND VALUES ------
                {data}
                ------- END COLUMN AND VALUES -------
                """

            human_msg = f"""{human_subprompt}

            Here are the target classes:
            ------ BEGIN TARGET CLASSES ------
            {targets}
            ------- END TARGET CLASSES -------"""

            prompt = [
                SystemMessage(content=sys_msg),
                HumanMessage(content=human_msg),
            ]

            return prompt

        def execute_prompt(prompt):
            try:
                if self.platform in ['openai', 'azure']:
                    with get_openai_callback() as cb:
                        response = self.llm(prompt)
                        response = response.content
                    stats = {
                        "total_tokens": cb.total_tokens,
                        "prompt_tokens": cb.prompt_tokens,
                        "completion_tokens": cb.completion_tokens,
                        "total_cost": round(cb.total_cost, 4)
                    }
                success = True
            except:
                success = False

            result = {
                "success": success,
                "response": response
            }

            return result

        # run the query
        result = {
            "success": False,
            "columns": {}
        }
        if use_content == True:
            for column, values in data.items():
                column_str = f"{column}: {', '.join(values)}"
                prompt = construct_prompt(column_str, targets, use_content)
                res = execute_prompt(prompt)
                if res.get("success"):
                    result["success"] = True
                    d = json.loads(res.get("response"))
                    key = list(d.keys())[0]
                    result['columns'][column] = d.get(key)
        else:
            columns = list(data.keys())
            prompt = construct_prompt(columns, targets, use_content)
            result = execute_prompt(prompt)
            result["success"] = True
            result['columns'] = json.loads(result.get("response"))
            d = result.pop("response")

        return result



class TestSADLClassifier(unittest.TestCase):
    def test_classification_1(self):
        data_mode = "csv"
        data_filepath = "data/Acme-Plan-Data.csv"
        entities_mode = "json"
        emap_file = "emap.json"

        labeller = SADLClassifier(data_mode=data_mode,
                                   data_filepath=data_filepath,
                                   context="",
                                   entities_filepath=emap_file)

        industry = labeller.classify_industry()
        success, result_type, result = labeller.classify_columns(context=industry)

        print("result_type: ", result_type)
        print("type: ", type(result))
        print("###################")
        self.assertTrue(success)
        self.assertIsInstance(result_type, str)
        if result_type == 'str':
            self.assertIsInstance(result, str)
        else:
            self.assertIsInstance(result, dict)

    def test_classification_2(self):
        data_mode = "df"
        data_filepath = "data/Acme-Plan-Data.csv"
        data = pd.read_csv(data_filepath)
        entities_mode = "json"
        emap_file = "emap.json"

        labeller = SADLClassifier(data_mode=data_mode,
                                   data=data,
                                   context="",
                                   entities_filepath=emap_file)

        industry = labeller.classify_industry()
        success, result_type, result = labeller.classify_columns(context=industry)

        print("result_type: ", result_type)
        print("type: ", type(result))
        print("###################")
        self.assertTrue(success)
        self.assertIsInstance(result_type, str)
        if result_type == 'str':
            self.assertIsInstance(result, str)  # Adjust this based on your expected result type
        else:
            self.assertIsInstance(result, dict)

if __name__ == '__main__':
    # # unit tests
    # unittest.main()

    # testbed
    csv = '...'
    targets = [
        "identifier",
        "label",
        "code",
        "gender",
        "datetime",
        "amount",
        "location",
        "category"
    ]
    labeller = SADLClassifier(platform="openai")
    data = labeller.load_data(csv, source="csv")
    mapping = labeller.map_to_targets(data, targets, use_content=True)
    print (json.dumps(mapping, indent=4))



import pytest
from llmsdk.lib import defaults

# common init vars for LLMs

@pytest.fixture
def init_llm():
    # vars for LLM
    params = {
        "agentname": "test_agent",
        "platform": defaults.LLM_PLATFORM,
        "model": defaults.LLM_MODEL,
        "embedding_model": defaults.LLM_EMBEDDING_MODEL,
    }
    return params



import os
import json
import pytest


class TestImport(object):
    """
    Test imports of modules
    """

    @pytest.mark.parametrize(
        "cls",
        [
            ("Node"),
            ("Transform"),
            ("Skin"),
            ("Compute"),
            ("Integration"),
            ("NotificationIntegration"),
            ("DatabaseIntegration"),
            ("SearchSkin"),
            ("GenericSkin"),
        ],
    )
    def test_cls(self, cls):
        """
        Test each cls
        """

        stmt = "from enrichsdk import {0}".format(cls)
        exec(stmt)

    def test_cls_list(self):
        """
        Test whether Node is capturing all classes
        """

        from enrichsdk import Node

        print(Node.implementation_list())



import json
import time
import string
from re import sub

from langchain_core.prompts import PromptTemplate
from langchain_core.messages import HumanMessage, SystemMessage
from langchain_community.callbacks.manager import get_openai_callback

from . import agent_events
from llmsdk.lib import defaults
from llmsdk.lib.enablers import AgentEnablersMixin
from llmsdk.agents.basellmrag import BaseLLMRAGAgent
from ..lib import extractors
from llmsdk.lib import SafeEncoder

__all__ = ['LLMMultiStepAgent']

class LLMMultiStepAgent(BaseLLMRAGAgent, AgentEnablersMixin):
    """
    Class to do multi-step actions using LLMs
    Step can include RAG, simple prompting, tool use, and more
    """

    def __init__(self,
                 name,
                 cred={},
                 platform=defaults.LLM_PLATFORM,
                 model=defaults.LLM_MODEL,
                 embedding_model=defaults.LLM_EMBEDDING_MODEL,
                 searchapi=defaults.LLM_SEARCH_API,
                 statestore=defaults.LLM_STATE_STORE,
                 topk=7,
                 chunk_size=1000,
                 chunk_overlap=300):
        """
        init the LLM query agent
        name: name of the agent
        cred: credentials object
        platform: name of the LLM platform backend to use
                default to OpenAI GPT platform for now, Azure is also supported
                will be extended in the future to suuport other models
        memory_size: how many tokens of memory to use when chatting with the LLM
        """

        start_time = time.time()

        # init the base class
        super().__init__(name=name,
                         cred=cred,
                         platform=platform,
                         model=model,
                         embedding_model=embedding_model,
                         agent_type="multistep",
                         searchapi=searchapi,
                         statestore=statestore)

        # defaults
        self.chunk_size     = chunk_size
        self.chunk_overlap  = chunk_overlap
        self.index          = None
        self.metadata       = {}
        self.vdb_client     = None
        self.index_name     = None
        self.index_store    = None
        self.topk           = topk
        self.doc_signatures = []
        self.docs           = {}

        # LLM params
        self.platform           = platform
        self.model              = model
        self.searchapi          = searchapi
        self.embedding_model    = embedding_model

        # init the llm and embeddings objects
        self.llm, self.embeddings = self._get_llm_objs(platform=self.platform,
                                                        model=self.model,
                                                        embedding_model=self.embedding_model,
                                                        cred=self.cred)

        # init the agent for searches
        self.llm_agent_srch, self.searchengine = self._load_search_agent(cred=self.cred,
                                                                          searchapi=self.searchapi,
                                                                          llm=self.llm)

        # note metadata for this agent
        self.metadata = {
            "agent": {
                "name": self.agent_name,
                "name": self.agent_type,
                "platform": self.platform,
                "model": self.model,
                "embedding_model": self.embedding_model,
            },
            "events": []
        }
        # log that the agent is ready
        duration = time.time() - start_time
        event = self._log_event(agent_events._EVNT_READY, duration)

    ## helper functions

    def _get_query_prompt_internal(self, docs, query, context, instructions):
        """
        generate a prompt for running a query in internal mode
        """
        sys_msg = """You are a highly advanced AI program designed to extract specific pieces of information from business documents.
Be precise in the responses you provide. Do not respond in full sentences, but only with the specific information asked for.
You will be provided with relevant pieces of context extracted from a business document to answer the question at the end.
If the question asks you to format your answer in some specific manner, do so.
If the question includes instructions to help you with extraction, follow those instructions.
If you cannot find the answer in the context, just say 'unknown', don't try to make up an answer and do not provide any explanations.
"""
        docs = [d.page_content for d in docs]

        human_msg = f"""
------ BEGIN DOCUMENT CONTEXT ------
{docs}
------ END DOCUMENT CONTEXT ------

------ BEGIN QUESTION ------
{context}
{query}
{instructions}
------ END QUESTION ------

Your response:"""

        messages = [
            SystemMessage(content=sys_msg),
            HumanMessage(content=human_msg),
        ]

        return messages

    def _get_query_prompt_norag(self, query, context, instructions):
        """
        generate a prompt for running a query without doing RAG
        """
        sys_msg = f"""You are a highly advanced AI program designed answer questions in natural language.
Your task is to read some text that is given to you and answer all questions about that text.
If you are provided with context to help you answer the question, pay attention to the context.
If the question asks you to format your answer in some specific manner, do so.
If the question includes instructions to help you with your task, follow those instructions exactly.
If you don't know the answer, just say 'unknown', don't try to make up an answer and do not provide any explanations.
"""

        human_msg = f"""
------ BEGIN CONTEXT ------
{context}
------ END CONTEXT ------

------ BEGIN QUESTION ------
{query}
{instructions}
------ END QUESTION ------

Your response:"""

        messages = [
            SystemMessage(content=sys_msg),
            HumanMessage(content=human_msg),
        ]

        return messages

    ## interfaces

    def read_document(self, source, content, metadata={}, params={}, store="chroma", persist_directory=None):
        """
        wrapper function that takes in the path to a document
        and sets it up for reading by the agent
        this function will create a new index if the agent does not already have one
        else it will use the existing index pointer
        needs the persist_directory that the index will use
        """
        # load the document
        data = self.load_data(source=source, content=content)

        # add the document to index
        if not self.index:
            # we have to init a new index
            self.create_add_index(data=data,
                                   store=store,
                                   persist_directory=persist_directory,
                                   index_name=self.agent_name)
        else:
            # we can use the agent's index pointer
            self.add_to_index(data)

        # extract text from document if it is a pdf
        # so that we have the table data
        if source in ["pdf"]:

            # run through Textract
            extracted_data = extract_text_from_file(content, provider="aws")

            # take the Textract output
            # and add tables and linetext to index
            for block in ["tables", "text"]:
                for extract in extracted_data:
                    # for each page in the document
                    for entry in extract[block]:
                        # for each table in the page
                        if any(f not in entry for f in ['id', 'content']):
                            continue
                        metadata = { "source": entry['id'] }
                        data = self.load_data(source="str",
                                               content=entry['content'],
                                               metadata=metadata)
                        self.add_to_index(data)

            # add the signature details to the agent's knowledge
            for extract in extracted_data:
                signatures = extract["signatures"]
                self.doc_signatures.extend(signatures)

        return

    def run_query_search(self, query, filters=None, context="", instructions=""):
        """
        run a query using the search agent
        this is useful when looking for answers using a search engine
        """
        def extract_content_sources(sourcedata):
            docs = sourcedata.get('organic_results')
            if not docs:
                return None
            sources = [{"content": d.get('snippet', ""), "source": d.get('link')} for d in docs]
            return sources

        # get the human-readable result
        input = {"input": query}
        response = self.llm_agent_srch.invoke(input)
        result = response.get("output", self._err_msg('search'))

        # get the sources
        sourcedata = self.searchengine.results(query)
        sources = extract_content_sources(sourcedata)
        if not sources:
            sources = [{"content": "", "source": f"search-{self.searchapi}"}]

        # construct result
        result = {
            "question": query,
            "answer": result,
            "suggest": list(set([q.get('question', '') for q in sourcedata.get('related_questions', [])])),
            "sources": sources
        }

        return result

    def run_query_internal(self, query, filters=None, context="", instructions=""):
        """
        run a query using llm on an internal docset indexed in index
        this is useful when looking for answers using a private source of data
        """
        # get the similar docs
        docs = self.get_similar_docs(query, filters=filters, topk=self.topk)

        # construct the prompt
        prompt = self._get_query_prompt_internal(docs, query, context, instructions)

        # call the LLM
        response = self.llm.invoke(prompt)

        # run the query against the similar docs
        result = {
            "question": query,
            "answer": response.content,
            "sources": [{"content": d.page_content, "metadata": d.metadata, "distance": d.metadata.pop('distance')} for d in docs],
        }

        return result

    def run_query_norag(self, query, filters=None, context="", instructions=""):
        """
        run a query using llm to answer questions about some text data
        this is useful when looking for answers about some context but without doing RAG
        """
        # construct the prompt
        prompt = self._get_query_prompt_norag(query, context, instructions)

        # call the LLM
        response = self.llm(prompt)

        result = {
            "question": query,
            "answer": response.content,
        }

        return result

    def query(self, query, filters=None, context="", instructions="", mode="internal"):
        """
        run a query on an index using an llm chain object
        query: query string
        mode: 'internal' for querying over docset, 'search' for searching the web
        """

        start_time = time.time()

        method = getattr(self, f"run_query_{mode}", None)
        if method is None:
            raise Exception(f"Unsupported mode: {mode}")

        try:
            if self.platform in ['openai', 'azure']:
                with get_openai_callback() as cb:
                    result = method(query, filters=filters, context=context, instructions=instructions)
                stats = {
                    "total_tokens": cb.total_tokens,
                    "prompt_tokens": cb.prompt_tokens,
                    "completion_tokens": cb.completion_tokens,
                    "total_cost": round(cb.total_cost, 4)
                }
            else:
                result = method(query, filters=filters, context=context, instructions=instructions)
                stats = {}
        except:
            result = {
                "question": query,
                "answer": self._err_msg('field'),
                "sources": [],
            }
            stats = {}

        # log the event
        params = {
            "query": query,
            "mode": mode,
            "result": result.copy() if result is not None else None,
            "stats": stats,
        }
        duration = time.time() - start_time
        event = self._log_event(agent_events._EVNT_QUERY, duration, params=params)

        # add the event to the result
        result['metadata'] = {
            "timestamp": event['timestamp'],
            "duration": event['duration'],
        }

        return result

    def process_spec_queries(self, spec):
        """
        take a spec containing questions and answer them
        against the docset indexed by the agent
        """

        def get_depend_params(depends, step_answers, query_answers):
            params = {}
            stop_steps = False

            for p in depends.get("steps", []):
                a = step_answers.get(p, "")
                if len(a) == 0:
                    stop_steps = True
                params[p] = a
            for p in depends.get("queries", []):
                a = query_answers.get(p, "")
                if len(a) == 0:
                    stop_steps = True
                params[p] = a

            return params, stop_steps

        # retry answers
        retry_answers = ["im having trouble understanding try another way of wording your query",
                         "unknown"]

        # get the list of queries
        query_set = spec.get("query_set", [])
        indx_query_set = {}

        # begin an empty dict
        extracted_info = {}
        query_answers = {}
        grounding = {}

        # spec name
        spec_name = spec.get("name", "unknown")

        # foreach query to process
        for one_query in query_set:

            step_answers = {}

            enable = one_query.get("enable", True)
            q_name = one_query['name']

            # make note of this one_query in the index
            # we'll need this later when re-ordering
            indx_query_set[q_name] = one_query

            if not enable:
                self.logger.debug(f"Skipping query: {q_name} [enable=false]",
                                     extra={
                                         'source': self.agent_name,
                                         'data': json.dumps(one_query, indent=4, cls=SafeEncoder)
                                     })
                continue

            self.logger.debug(f"Running query: {q_name}",
                                 extra={
                                     'source': self.agent_name,
                                     'data': json.dumps(one_query, indent=4, cls=SafeEncoder)
                                 })

            # get the set of steps for this query
            steps = one_query.get("steps", [])

            # for each step
            for one_step in steps:

                # per query vars
                step_name = one_step['name']
                query_mode = one_step.get("mode", "internal")
                query = one_step.get("query")
                filters = one_step.get("filters")
                context = one_step.get("context", "")
                instructions = one_step.get("instructions", "")
                query_alts = one_step.get("query_alts", [])
                use_alts = one_step.get("use_alts", "on-fail")
                depends = one_step.get("depends", {})
                postprocess = one_step.get("postprocess", {})

                if query_mode == "callback":
                    # this is a special case
                    # no calls go out to the LLM
                    # instead a custom callback handler will deal with this step
                    params, stop_steps = get_depend_params(depends, step_answers, query_answers)
                    if stop_steps:
                        break

                    handler = one_step.get("handler")
                    if callable(handler):
                        answer = handler(params)
                    else:
                        self.logger.error(f"Invalid handler for '{query_mode}' mode in step: {step_name}",
                                             extra={
                                                 'source': self.agent_name,
                                             })
                        break

                    # note the answer
                    # we may need it for subs in later queries
                    step_answers[step_name] = answer
                    query_answers[q_name] = answer

                    # add the answers to the columns in the extracted dataset
                    ans = extracted_info.get(q_name,[])
                    ans.append(answer)
                    extracted_info[q_name] = ans

                    # move on to the next step
                    continue

                if not query:
                    continue

                # post-processing handling
                pp_handler = postprocess.get("handler")
                pp_response = postprocess.get("response", "fill")
                pp_othercols = postprocess.get("othercols", [])

                # collect all the queries we need to run
                queries = [query] + query_alts

                # do any query modifications
                params, stop_steps = get_depend_params(depends, step_answers, query_answers)
                if stop_steps:
                    break
                # var substition for the queries
                queries = [q.format(**params) for q in queries]
                # var substition for the instructions
                instructions = instructions.format(**params)
                # var substition for the context
                context = context.format(**params)

                # run the queries against the LLM
                for q_cnt, query in enumerate(queries):

                    print (f"--> {q_name}:{step_name}:Q{q_cnt}:{query}")

                    # get the answer
                    response = self.query(query, filters=filters, context=context, instructions=instructions, mode=query_mode)
                    answer = response['answer']
                    sources = response.get('sources', [])
                    # normalized answer
                    s_answer = answer.translate(str.maketrans('', '', string.punctuation)).lower().strip()

                    # check the answer for UNKNOWN
                    # but only in the case when postprocess->response==fill
                    # this is for correct handling of spec extension/replacement
                    if pp_response == 'fill':
                        if s_answer in retry_answers:
                            continue

                    # post-process if needed
                    # check if handler is callable
                    if callable(pp_handler):
                        # collect all the columns needed to post-process the answer
                        params = {q_name: answer}
                        for col in pp_othercols:
                            if col in extracted_info:
                                if isinstance(extracted_info[col], list):
                                    val = extracted_info[col][0]
                                else:
                                    val = extracted_info[col]
                                params[col] = val
                        answer = pp_handler(params)

                    # note the answer
                    # we may need it for subs in later queries
                    step_answers[step_name] = answer
                    query_answers[q_name] = answer

                    # if next step action is 'extend'
                    # then, we need to extend the query spec
                    if pp_response == "extend":
                        query_set.extend(answer)
                        # continue, so that we move to the next query
                        continue

                    # if next step action is 'replace'
                    # then, we need to replace the query_set items in the default query spec
                    # with the newly loaded one
                    if pp_response == "replace":
                        # we have a new spec here to replace the current running spec
                        # get the spec name
                        spec_name = answer.get("name", spec_name)
                        # get the new query set
                        new_query_set = answer.get("query_set", [])
                        # get the names of all query items in replace spec
                        replace_query_set_items = {}
                        for q_item in new_query_set:
                            replace_query_set_items[q_item['name']] = q_item
                        # now, run through each query item in the default query set
                        # and check if replacement is needed
                        for i in range(len(query_set)):
                            if query_set[i]['name'] in replace_query_set_items:
                                # we have found a query item that needs to be replaced
                                query_set[i] = replace_query_set_items[query_set[i]['name']]

                        # continue, so that we move to the next query
                        continue

                    # add the answers to the columns in the extracted dataset
                    ans = extracted_info.get(q_name,[])
                    if answer in ans:
                        # we have found this answer before
                        # no need to collect it again
                        continue
                    ans.append(answer)
                    extracted_info[q_name] = ans

                    # at this point, at least one alt query has response
                    # collect the grounding elements
                    if len(sources) > 0:
                        curr_sources = grounding.get(q_name, [])
                        curr_sources.append(sources)
                        grounding[q_name] = curr_sources

                    # check if we need to run other alts
                    if use_alts == "on-fail":
                        # no need to run an alt query
                        # since we have atleast some response
                        break


        # get the order of fields to return
        reordered_query_set = []
        queryset_order = spec.get("order")
        if not queryset_order:
            # we don't have an explicit order provided
            # use the default ordering
            reordered_query_set = query_set
        else:
            for qso in queryset_order:
                if qso in indx_query_set:
                    one_query = indx_query_set[qso]
                    reordered_query_set.append(one_query)

        # check if all columns exist
        # and add the collected grounding
        extracts = {}
        for one_query in reordered_query_set:
            # check if we need to inlcude this field
            enable = one_query.get("enable", True)
            if not enable:
                continue
            include = one_query.get("include", True)
            if not include:
                continue

            # we need to include this field
            q_name = one_query['name']
            q_title = one_query.get("title", q_name)
            default = one_query.get("default")
            default = [[]] if default == None else [default]
            extracts[q_name] = {
                "spec": spec_name,
                "title": q_title,
                "n_steps": 0 if not extracted_info.get(q_name) else len(extracted_info[q_name]),
                "steps": extracted_info.get(q_name),
                "answer": extracted_info.get(q_name, default)[-1],
                "sources": grounding.get(q_name, []),
            }

        return extracts

    def process_spec_signatures(self, spec):
        """
        check if signatures are present
        """
        self.logger.debug("Detecting signatures...",
                             extra={'source': self.agent_name})

        if spec.get("detect_signatures", False) == False:
            return None

        pages = []
        confidence = 0
        if len(self.doc_signatures) > 0:
            for signature in self.doc_signatures:
                pages.append(signature['page'])
                confidence += signature['confidence']
            n_sigs = len(pages) # this is correct
            pages = list(set(pages))
            n_pages = len(pages)
            confidence = round(confidence/n_sigs, 2)

            comment = f"Detected {n_sigs} signatures across {n_pages} pages"

            signatures = {
                "found": True,
                "comment": comment,
                "n_signatures": n_sigs,
                "n_pages": n_pages,
                "pages": pages,
                "confidence": confidence,
            }
        else:
            signatures = {
                "found": False,
                "comment": f"No signatures detected",
            }

        return signatures

    def process_spec(self, spec):
        """
        process a profilespec
        """
        name = spec.get("name")
        self.logger.debug(f"Processing spec: {name}",
                             extra={
                                 'source': self.agent_name,
                                 'data': json.dumps(spec, indent=4, cls=SafeEncoder)
                             })

        # check for signatures
        signatures = self.process_spec_signatures(spec)

        # process the questions
        extracts = self.process_spec_queries(spec)

        result = {
            "spec": name,
            "timestamp": time.time(),
            "extracts": extracts,
            "signatures": signatures
        }

        return result

    def process_hierarchical_spec(self, spec):
        """
        Take a MultiStep agent and a set of specs and run through all the specs
        This is a hierarchical spec processor
        Inputs:
            specs -> dict of specs, each key should point to a spec that will be processed
            init_spec_name -> what spec key should we start the process at
            spec_iter -> name of field to iterate over keys in specs dict
        Return:
            extracts -> dict of extract objects keyed by spec name
        """

        init_spec = spec.get('init_spec')
        iter_key = spec.get('iter_key')
        specs = spec.get('specs')
        if init_spec == None or iter_key == None or specs == None:
            raise Exception(f"Malformed multi spec: {spec}")

        # get the starting spec
        one_spec = specs.get(init_spec)
        if one_spec == None:
            return None

        # process the starting spec
        info = self.process_spec(one_spec)

        # init the dict to collect all extract objects
        extracts = {
            init_spec: info
        }

        # process each section spec
        sections = json.loads(info['extracts'].get(iter_key, {}).get('answer', '{}'))
        for section, pages in sections.items():

            # get the spec for the section
            one_spec = specs.get(section)
            if one_spec == None:
                continue

            # process this spec
            name = one_spec.get("name", section)
            info = self.process_spec(one_spec)

            # add the spec name to the extracts
            for col, details in info['extracts'].items():
                info['extracts'][col]["spec"] = name

            # collect this set of extract
            extracts[name] = info

        return extracts

    def extracts_to_df(self, info):
        """
        Convert extracts object into dataframe
        """
        extracts = info['extracts']
        entries = []

        for col, details in extracts.items():

            sources = []
            if len(details['sources'])>0:
                for sourceset in details['sources']:
                    for source in [sourceset[0]]:
                        sources.append(source['content'])

            steps = []
            steps = [] if not isinstance(details['steps'], list) else details['steps']

            entry = {
                "field": col,
                "title": details['title'],
                "answer": details['answer'],
                "grounding": "\n\n".join(sources),
                "n_steps": details['n_steps'],
                "steps": "\n".join(steps),
            }
            entries.append(entry)

        df = pd.DataFrame(entries)

        return df


################### BEGIN TESTBED ###################

if __name__ == "__main__":

    # vars
    # cred = get_credentials_by_name('openai-api')
    persist_directory = "chromadb123"
    path = '...'

    agentname = "test_agent"
    # profilespec = get_profilespec()
    print(json.dumps(profilespec, indent=4, cls=SafeEncoder))
    platform = "azure"

    # create an agent
    agent = LLMMultiStepAgent(name=agentname, platform=platform)

    # load the data
    data = agent.load_data(content=path)

    # add to index
    agent.create_add_index(data=data,
                           store="chroma",
                           persist_directory=persist_directory,
                           index_name=agentname)

    # run the profilespec of queries
    info = agent.process_spec(profilespec)

    # output extraction
    for field, extract in info.get("extracts", {}).items():
        print (f"{field}: {extract.get('answer')}")



#!/usr/bin/env python3
"""

Script to extract tables from pdfs or image files. This will be piped through the textract

"""
import os
import sys
import json
import tempfile
import traceback
import re
import inspect
import importlib
import csv
from collections import defaultdict

import click
import requests

from asgiref.sync import async_to_sync, sync_to_async

from langchain_community.vectorstores import Chroma
from langchain_community.embeddings import OpenAIEmbeddings

import llmsdk
from llmsdk.lib import *
from llmsdk.agents import *
from llmsdk.services.lib import get_files, get_agent_details
from llmsdk.services.qna import get_qna_agent, supported_exts as qna_supported_exts
from llmsdk.services.docgen import get_docgen_agent, supported_exts as docgen_supported_exts
from llmsdk.services.datagpt import get_datagpt_agent, supported_exts as datagpt_supported_exts
from llmsdk.services.extractor import get_contract_profilespec, contract_query

from llmsdk.lib.summarizers import *

#####################################
# Helper Functions...
#####################################
def get_chroma_index_summary_generic(indexname, persistdirectory):

    db = Chroma(indexname,
                embedding_function=OpenAIEmbeddings(),
                persist_directory=persistdirectory)

    content = db.get()

    summary = {
        'count': 0,
        'chunks': defaultdict(int)
    }
    metadatas = content['metadatas']
    for entry in metadatas:
        summary['count'] += 1
        if 'sample' not in summary:
            summary['sample' ] = entry
        #chunk = entry['chunk']
        #summary['chunks'][chunk] += 1

    return summary

def get_chroma_index_summary_files(indexname, persistdirectory):

    db = Chroma(indexname,
                embedding_function=OpenAIEmbeddings(),
                persist_directory=persistdirectory)

    content = db.get()

    sample = {
        "source": "/home/pingali/Data/enrich/llm-agents/data/docs/qna-batch-brookfield/October Three RFP/converted/Ste Michelle Bid Specifications 06.21.2023.pdf",
        "file_path": "/home/pingali/Data/enrich/llm-agents/data/docs/qna-batch-brookfield/October Three RFP/converted/Ste Michelle Bid Specifications 06.21.2023.pdf",
        "page": 0,
        "total_pages": 7,
        "format": "PDF 1.5",
        "title": "",
        "author": "Emmy Brown",
        "subject": "",
        "keywords": "",
        "creator": "Writer",
        "producer": "LibreOffice 6.4",
        "creationDate": "D:20230719202934+05'30'",
        "modDate": "",
        "trapped": "",
        "bboxes": "[[264.3999938964844, ...,489.29425048828125]]",
        "file": "Ste Michelle Bid Specifications 06.21.2023.pdf",
        "chunk": 1,
        "id": "922e098a97dd5fcea7343d052317ee71"
        }

    summary = {}
    metadatas = content['metadatas']
    for entry in metadatas:
        filename = entry['file']
        page     = entry['page']
        chunk    = entry['chunk']
        format_  = entry['format']
        if filename not in summary:
            summary[filename] = {
                'pages': [],
                'chunks': [],
                'format': "unknown"
            }
        if page not in summary[filename]['pages']:
            summary[filename]['pages'].append(page)
        summary[filename]['chunks'].append(chunk)
        summary[filename]['format'] = format_

    return summary

@click.group()
def process():
    """
    Run the agent in batch mode with documents
    """

def show_help():
    helptext = """This script needs to run with

export DATA_ROOT="/home/pingali/Data/enrich/llm-agents/data/docs"
export AGENTNAME=default
    """
    print(helptext)


#####################################
# QnA Simple - single command
#####################################
@process.command("qna-simple")
@click.argument("dataset")
@click.argument("question")
def qna(dataset, question):
    """
    Ask a question against a qna agent
    """

    if 'DATA_ROOT' not in os.environ:
        show_help()
        return

    details = async_to_sync(get_agent_details)(namespace="qna",
                                               username="cli",
                                               subdir=dataset,
                                               exts=qna_supported_exts,
                                               get_task_specific_agent=get_qna_agent)

    result = details['agent'].query(question, mode="internal")
    print(json.dumps(result, indent=4))

#####################################
# qna-batch execution
#####################################
def get_path_source(fpath):
    if os.path.isdir(fpath):
        return "dir"
    else:
        return fpath.split('.')[-1]

@process.command("qna-batch")
@click.option("--input", "-i",
              required=True,
              multiple=True,
              help="Path to input docset/file")
@click.option("--queryspec", "-q",
              required=True,
              help="Path to queryspec")
@click.option("--output", "-o",
              required=True,
              help="Path to store output json")
@click.option("--platform", "-p",
              default="openai",
              help="LLM platform backend to use {openai, azure}")
@click.option("--model", "-m",
              default="gpt-4o-mini",
              help="LLM model to use {gpt-3.5-turbo, gpt-4o-mini, gpt-4o}")
@click.option("--agenttype", "-t",
              default="queryextract",
              help="Agent type")
@click.option("--agentname", "-a",
              required=True,
              help="Agent name")
@click.option("--indexname", "-x",
              required=True,
              help="Name of index")
@click.option("--force-reindexing/--no-force-reindexing",
              required=False,
              default=True,
              help="Whether the index should be built")
@click.option("--vectorstore", "-s",
              default="chroma",
              help="Flavour of Vector DB (chroma, faiss)")
@click.option("--persistdirectory", "-p",
              default="chromadb",
              help="Path to directory to persist vector DB")
def qna_batch(agenttype, agentname, indexname,
              input, queryspec,
              output, platform, model, vectorstore,
              persistdirectory, force_reindexing):
    """
    Take a query spec and docset and extract information in batch mode
    """

    # create an agent
    if agenttype == 'queryextract':
        agent = LLMQuerierExtractor(name=agentname,
                                    platform=platform)
    elif agenttype in ['multistep']:
        agent = LLMMultiStepAgent(name=agentname,
                                    platform=platform, model=model)
    elif agenttype in ['hierarchical']:
        agent = LLMMultiStepAgent(name=agentname,
                                    platform=platform, model=model,
                                    chunk_size=5000, chunk_overlap=1000)
    else:
        # default to the older qna agent
        agent = LLMQuerierExtractor(name=agentname, platform=platform)


    # Cleanup the name. max 60 chars, starts and ends with
    # alphanumeric char
    indexname = slugify(indexname)[:60]
    indexname = re.sub(r"^[^a-zA-Z0-9]+","", indexname)
    indexname = re.sub(r"[^a-zA-Z0-9]+$","", indexname)

    print("Index", indexname)
    print("Persist Directory", persistdirectory)
    print("Force reindexing", force_reindexing)

    try:

        stats = None
        create = True

        if not force_reindexing:
            try:
                print("Trying to load index")
                agent.load_index(store=vectorstore,
                                 persist_directory=persistdirectory,
                                 index_name=indexname)
                stats = agent.get_index_stats()
                create = False
            except:
                #traceback.print_exc()
                pass
    except:
        # traceback.print_exc()
        stats = None
        create = True

    if ((stats is not None) and (stats['n_items'] == 0)):
        create = True

    if not create:
        summary = get_chroma_index_summary_files(indexname, persistdirectory)
        print(json.dumps(summary, indent=4))

    if create:
        print("Creating index")
        cnt = 0
        for fpath in input:
            # figure out the source
            source = get_path_source(fpath)

            # load the data
            data = agent.load_data(source=source, content=fpath)

            # add to index
            if cnt == 0:
                # only for the first file
                agent.create_add_index(data=data,
                                       store=vectorstore,
                                       persist_directory=persistdirectory,
                                       index_name=indexname)
            else:
                # for subsequent files
                agent.add_to_index(data=data)

            cnt += 1

    # load query spec
    queryspecfile = queryspec # dont confuse spec with filename of that spec
    if queryspecfile.endswith(".json"):
        with open(queryspecfile, "r") as fd:
            queryspec = json.load(fd)
    elif queryspecfile.endswith(".py"):
        module_name = "queryspecmod"
        spec = importlib.util.spec_from_file_location(module_name, queryspec)
        module = importlib.util.module_from_spec(spec)
        sys.modules[module_name] = module
        spec.loader.exec_module(module)

        # Now get hold of query spec
        queryspec = module.get_queryspec()
    else:
        raise Exception(f"Unable to load the query spec: {queryspecfile}")

    # Limit the questions...for testing..
    # queryspec['query_set'] = queryspec['query_set'][:1]

    # process the spec
    if agenttype in ['queryextract', 'multistep']:
        responses = agent.process_spec(queryspec)
    elif agenttype in ['hierarchical']:
        responses = agent.process_hierarchical_spec(queryspec)
    else:
        responses = agent.process_spec(queryspec)

    # store
    with open(output, "w") as fd:
        json.dump(responses, fd, indent=2)
    print (f"Stored response json at: {output}")

#####################################
# news summarizer execution
#####################################
@process.command("news-summarizer")
@click.option("--agentname", "-a",
              required=True,
              help="Agent name")
@click.option("--action", "-c",
              default="summarize",
              help="Action to perform on news results {summarize, dedup}")
@click.option("--input", "-i",
              required=True,
              help="Path to news results")
@click.option("--rulespec", "-r",
              help="Path to rulespec")
@click.option("--output", "-o",
              required=True,
              help="Path to store output json")
@click.option("--platform", "-p",
              default="openai",
              help="LLM platform backend to use {openai, azure}")
@click.option("--model", "-m",
              default="gpt-4o-mini",
              help="LLM model to use {gpt-3.5-turbo, gpt-4o-mini, gpt-4o}")
def news_summarizer(agentname,
                      action, input, rulespec, output,
                      platform, model):
    """
    Take a summarization rule spec, news results json,
    and execute against a generic prompt LLM agent
    """

    # create a summarizer object
    summarizer = GoogleNewsSummarizer()
    # create an prompter agent
    # we need to pass this agent to the summarizer object
    agent = LLMPromptAgent(name=agentname,
                            platform=platform,
                            model=model)

    # load the news results data
    serp_results = summarizer.load_serp_results(input)

    if action == "summarize":
        # load rule spec
        rulespecfile = rulespec # dont confuse spec with filename of that spec
        if rulespecfile and rulespecfile.endswith(".py"):
            module_name = "queryspecmod"
            spec = importlib.util.spec_from_file_location(module_name, rulespec)
            module = importlib.util.module_from_spec(spec)
            sys.modules[module_name] = module
            spec.loader.exec_module(module)

            # Now get hold of rules spec
            rulespec = module.get_rulespec()
        else:
            raise Exception(f"Unable to load the rule spec: {rulespecfile}")

        # run the summarization
        serp_results = summarizer.process_results(agent, serp_results, rulespec)
    else:
        # run the dedup
        serp_results = summarizer.dedup_results(agent, serp_results)

    # store the results
    summarizer.store_results(serp_results, output)
    print (f"Stored response json at: {output}")

#####################################
# dataset summarizer
#####################################
@process.command("dataset-summarizer")
@click.option("--agentname", "-a",
              required=True,
              help="Agent name")
@click.option("--input", "-i",
              required=True,
              help="Path to csv")
@click.option("--rulespec", "-r",
              required=True,
              help="Path to rulespec")
@click.option("--output", "-o",
              required=True,
              help="Path to store output json")
@click.option("--platform", "-p",
              default="openai",
              help="LLM platform backend to use {openai, azure}")
@click.option("--model", "-m",
              default="gpt-4o-mini",
              help="LLM model to use {gpt-3.5-turbo, gpt-4o-mini, gpt-4o}")
def dataset_summarizer(agentname,
                       input, rulespec, output,
                       platform, model):
    """
    Take a summarization rule spec, dataset csv
    and execute against a generic prompt LLM agent
    """

    # load prompt spec
    rulespecfile = rulespec # dont confuse spec with filename of that spec
    if rulespecfile.endswith(".py"):
        module_name = "queryspecmod"
        spec = importlib.util.spec_from_file_location(module_name, rulespec)
        module = importlib.util.module_from_spec(spec)
        sys.modules[module_name] = module
        spec.loader.exec_module(module)

        # Now get hold of rules spec
        rulespec = module.get_rulespec()
    else:
        raise Exception(f"Unable to load the rule spec: {rulespecfile}")

    # create a summarizer object
    summarizer = DatasetSummarizer()

    # create an prompter agent
    # we need to pass this agent to the summarizer object
    agent = LLMPromptAgent(name=agentname,
                           platform=platform,
                           model=model)
    # load the news results data
    data = summarizer.load_dataset(input)

    # run the summarization
    result = summarizer.process_results(agent, data, rulespec)

    # store the results
    summarizer.store_results(result, output)
    print (f"Stored response json at: {output}")


#####################################
# DocGen
#####################################
@process.command("docgen-run")
@click.argument("dataset")
@click.argument("specification")
@click.option("--output", "-o",
              help="Output file")
def docgen(dataset, specification, output):
    """
    Generate document
    """

    if 'DATA_ROOT' not in os.environ:
        show_help()
        return

    # Load the specification
    spec = json.load(open(specification))

    # Get the agent
    details = async_to_sync(get_agent_details)(namespace="docgen",
                                               username="cli",
                                               subdir=dataset,
                                               exts=docgen_supported_exts,
                                               get_task_specific_agent=get_docgen_agent)

    # Generate the output
    answer   = details['agent'].generate_doc(spec)

    # => Now output the result..
    text = answer # answer not a json
    if output is not None:
        output = os.path.abspath(output)
        with open(output, 'w') as fd:
            fd.write(text)
        print(f"Output: {output}")
    else:
        print(text)

@process.command("docgen-sample-spec")
@click.option("--output", "-o",
              help="Output file")
def docgen_spec(output):
    """
    Generate a sample specification
    """
    profilespec = {
	    "prompt": {
		    "prefix": "write a para about",
		    "suffix": "for this startup company"
	    },
	    "sections": {
		    "about": {
			    "enable": True,
		        "title": "About the Company",
			    "intent": "introducing the startup",
			    "questions": [
				    "industry and specific focus",
				    "key technologies developed",
				    "unique value proposition"
			    ],
			    "collate": False,
		    },
		    "founders": {
			    "enable": True,
			    "title": "Company Founders",
			    "intent": "about the co-founders of the startup",
			    "questions": [
				    "co-founders and the roles they hold",
				    "past experience of co-founders"
			    ],
			    "collate": "rewrite"
		    },
		    "market-analysis": {
			    "enable": True,
			    "title": "Market Analysis",
			    "intent": "about market analysis for this startup",
		        "questions": [
			        "target customer base",
			        "other potential customers",
				    "other comparable companies and startups in this space",
				    "the main competitors",
				    "biggest risk factors",
				    "key differentiators",
				    "market trends for the domain"
			    ],
			    "collate": False
		    }
	    }
    }

    text = json.dumps(profilespec, indent=4)
    if output is not None:
        output = os.path.abspath(output)
        with open(output, 'w') as fd:
            fd.write(text)
        print(f"Output: {output}")
    else:
        print(text)

#####################################
# DataGPT
#####################################
@process.command("datagpt")
@click.argument("dataset")
@click.argument("question")
@click.option("--output", "-o",
              default=None,
              help="Output file")
def datagpt(dataset, question, output):
    """
    Interact with a dataset
    """

    if 'DATA_ROOT' not in os.environ:
        show_help()
        return

    details = async_to_sync(get_agent_details)(namespace="datagpt",
                                               username="cli",
                                               subdir=dataset,
                                               exts=datagpt_supported_exts,
                                               get_task_specific_agent=get_datagpt_agent)

    result = details['agent'].query(question)

    text = json.dumps(result, indent=4)
    if output is not None:
        output = os.path.abspath(output)
        with open(output, 'w') as fd:
            fd.write(text)
        print(f"Output: {output}")
    else:
        print(text)

#####################################
# DATA CLEANER (using SADL)
#####################################
@process.command("data-cleaner")
@click.option("--platform", "-p",
              default="openai",
              help="LLM platform backend to use {openai, azure}")
@click.option("--action", "-a",
              default="map",
              help="Type of action to perform {map, label}")
@click.option("--input", "-i",
              required=True,
              help="Path to input CSV")
@click.option("--targets", "-t",
              default=None,
              help="Path to target classes CSV")
@click.option("--usecontent", "-c",
              default=False,
              help="Whether to use data samples while labelling")
@click.option("--output", "-o",
              default=None,
              help="Output file")
def data_cleaner(platform, action, input, targets, usecontent, output):
    """
    Take a CSV file and label the columns
    Two modes are supported via the action param
        - map: map each column name to a provided list of target classes
        - label: try to guess the column label automatically [not supported right now]
    """

    # init the agent
    labeller = SADLClassifier(platform=platform)

    # check for the input file
    if not os.path.exists(input):
        print("Input file does not exist. Exiting.")
        return

    # load the data
    data = labeller.load_data(input, source="csv")

    if action == "map":
        # check for the targets file
        if not os.path.exists(targets):
            print("Target Classes file does not exist. Exiting.")
            return
        with open(targets, "r") as fd:
            target_classes = fd.readlines()
        # run the mapping
        mapping = labeller.map_to_targets(data, target_classes, use_content=usecontent)

    # store
    with open(output, "w") as fd:
        json.dump(mapping, fd, indent=2)
    print (f"Stored mapping json at: {output}")



#####################################
# Contract processing
#####################################
@process.command("contract-sample-spec")
@click.argument("specfile")
@click.option("--nature", "-n",
              default="contract",
              type=click.Choice(['contract'],
                                case_sensitive=False))
def contract_sample_spec(nature, specfile):
    """
    Get a sample spec for extractor

    Store as llm-runner extractor-sample-spec -n contract 2> .../work/spec.py
    """

    if 'DATA_ROOT' not in os.environ:
        show_help()
        return

    if os.path.exists(specfile):
        print("Specfile exists. Please remove first")
        print(specfile)
        return

    spec = inspect.getsource(get_contract_profilespec)
    spec += "\nget_spec = get_contract_profilespec\n"

    with open(specfile, 'w') as fd:
        fd.write(spec)

    print("Use this with extractor-run")
    print(f"llm-runner extractor-run {specfile} <source-document>")

@process.command("contract-run")
@click.argument("spec")
@click.argument("filename")
@click.option("--output", "-o",
              default=None,
              help="Output file")
def contract_extractor_run(spec, filename, output):
    """
    Run a extractor spec against the input files

    spec is generated using extractor-sample-spec
    """

    if ((os.path.abspath(filename) != filename) or
        (os.path.abspath(spec) != spec) or
        ((output is not None) and
         (os.path.abspath(output) != output))):
        print("This command required absolute paths for spec/inputfile/output")
        return

    if not os.path.exists(filename):
        print(f"Input file is missing: {filename}")
        return

    if not os.path.exists(spec):
        print(f"Profile spec file is missing: {spec}")
        return

    # Load the specification
    modspec = importlib.util.spec_from_file_location("samplemod", spec)
    mod = importlib.util.module_from_spec(modspec)
    modspec.loader.exec_module(mod)
    profilespec = mod.get_spec()

    result, metadata = contract_query(profilespec, filename)

    # output extraction
    text = json.dumps({
        "result": result,
        "metadata": metadata
    }, indent=4)

    if output is not None:
        output = os.path.abspath(output)
        with open(output, 'w') as fd:
            fd.write(text)
        print(f"Output: {output}")
    else:
        print(text)


@process.command("check-chroma-index")
@click.option("--persistdirectory", "-p",
              default="chromadb",
              help="Path to directory to persist vector DB")
@click.option("--indexname", "-x",
              required=True,
              help="Name of index")
@click.option('--nature', '-n',
              type=click.Choice(['files', 'generic']),
              default='generic')
@click.option("--output", "-o",
              required=False,
              help="Output File")
def check_chromadb(persistdirectory, indexname, output, nature):
    """
    Check vector db
    """
    indexname = slugify(indexname)
    if nature == 'files':
        summary = get_chroma_index_summary_files(indexname, persistdirectory)
    else:
        summary = get_chroma_index_summary_generic(indexname, persistdirectory)

    if output is None:
        print(json.dumps(summary, indent=4))
    else:
        with open(output, 'w') as fd:
            fd.write(json.dumps(summary, indent=4))

@click.group()
@click.option("--server", "-s",
              default="http://localhost:9010",
              help="Server")
@click.pass_context
def databank(ctx, server):
    """
    Databank commands
    """
    ctx.ensure_object(dict)
    ctx.obj = {
        "server": server
    }


@databank.command("health")
@click.pass_context
def databank_health(ctx):
    """
    Check databank health
    """

    server = ctx.obj['server']
    url = server + "/health"
    response = requests.get(url)
    response = response.json()
    print(json.dumps(response, indent=4))

@databank.command("load")
@click.argument("csvfile")
@click.argument("columns")
@click.pass_context
def databank_load_data(ctx, csvfile, columns):
    """
    Load data in databank
    """
    assert csvfile.endswith("csv")
    assert isinstance(columns, str) and len(columns) > 0

    columns = columns.split(",")
    reader = csv.DictReader(open(csvfile))
    data = list(reader)
    print(f"Read file with {len(data)} records")
    print(json.dumps(data[:2], indent=4))
    record = data[0]
    missing = [c for c in columns if c not in record]
    if len(missing) > 0:
        print(f"Columns not present in csv: {missing}")
        return
    print(f"All columns present: {columns}")

    server = ctx.obj['server']
    url = server + "/load"
    response = requests.post(url, json={
        "data": data,
        "params": {
            "index_columns": columns
        }
    })
    print("Response status", response.status_code)
    if response.status_code < 400:
        response = response.json()
        print(json.dumps(response, indent=4))
    else:
        print(response.text)

@databank.command("query")
@click.argument("query")
@click.pass_context
def databank_query(ctx, query):
    """
    Query databank
    """
    server = ctx.obj['server']
    url = server + "/qna"
    response = requests.post(url,
                             params={
                                 "query": query,
                                 "dataset": "default",
                                 "user": "anyone",
                             },
                             json={
                                 "policy": {},
                                 "extra": {}
                             })


    print("=> Query")
    print("Response status", response.status_code)
    if response.status_code < 400:
        response = response.json()
        print(json.dumps(response, indent=4))
        request_id = response['request_id']
    else:
        print(response.text)
        return

    url = server + "/qna/status"
    while True:
        print(f"=> Status {request_id}")
        response = requests.get(url, params={
            'request_id': request_id
        })
        if response.status_code < 400:
            response = response.json()
            print(json.dumps(response, indent=4))
            status = response['status']
            if status == 'pending':
                time.sleep(5)
                continue
            else:
                break
        else:
            print(response.text)
            break

process.add_command(databank)

def main():
    process()

if __name__ == "__main__":
    main()



import os
import json
import pytest

from enrichsdk.quality import *


def test_expectation_manager():
    """
    Test manager configuration
    """

    mgr = ExpectationManagerBase()
    with pytest.raises(NotInitialized) as excinfo:
        mgr.validate(None, None)

    mgr.initialize()

    with pytest.raises(InvalidExpectation) as excinfo:
        mgr.validate(None, None)

    with pytest.raises(NoExpectations) as excinfo:
        mgr.validate(None, [])


def test_unknown_expectation():
    """
    Test unknown expectation
    """

    mgr = ExpectationManagerBase()
    mgr.initialize()

    with pytest.raises(UnsupportedExpectation) as excinfo:
        mgr.validate(None, {"expectation": "Unknown_XXX"})


@pytest.mark.parametrize("df", [None, "", 1, []])
def test_table_columns_invalid_frame(df):
    """
    Test unknown expectation
    """

    mgr = ExpectationManagerBase()
    mgr.initialize()

    with pytest.raises(InvalidExpectation) as excinfo:
        mgr.validate(df, {"expectation": "table_columns_exist"})

    result = mgr.validate(
        df, {"expectation": "table_columns_exist", "params": {"columns": "hello"}}
    )

    assert len(result) == 1
    assert result[0]["expectation"] == "table_columns_exist"
    assert not result[0]["passed"]
    assert "extra" in result[0] and "Not a dataframe" in result[0]["extra"]["reason"]


@pytest.mark.parametrize(
    "df",
    [
        pd.DataFrame([]),
        pd.DataFrame([{"hello1": 1}]),
    ],
)
def test_table_columns_missing_columns(df):
    """
    Test unknown expectation
    """

    mgr = ExpectationManagerBase()
    mgr.initialize()

    with pytest.raises(InvalidExpectation) as excinfo:
        mgr.validate(df, {"expectation": "table_columns_exist"})

    result = mgr.validate(
        df, {"expectation": "table_columns_exist", "params": {"columns": "hello"}}
    )

    assert len(result) == 1
    assert result[0]["expectation"] == "table_columns_exist"
    assert not result[0]["passed"]
    assert "extra" in result[0] and "Missing column" in result[0]["extra"]["reason"]


@pytest.mark.parametrize(
    "df,error",
    [
        (pd.DataFrame([]), "Wrong position"),
        (pd.DataFrame([{"col1": 1}]), "Wrong position"),
        (pd.DataFrame([[1]], columns=["col1"]), "Wrong position"),
        (pd.DataFrame([[1, 2]], columns=["col1", "col2"]), "Wrong position"),
        (pd.DataFrame([[1, 2, 3]], columns=["col2", "col3", "col1"]), "Wrong position"),
    ],
)
def test_table_columns_with_position(df, error):
    """
    Test column positioning
    """

    mgr = ExpectationManagerBase()
    mgr.initialize()

    with pytest.raises(InvalidExpectation) as excinfo:
        mgr.validate(df, {"expectation": "table_columns_exist"})

    # Test position = 1
    result = mgr.validate(
        df,
        {
            "expectation": "table_columns_exist_with_position",
            "params": {"columns": {"hello": 1}},
        },
    )

    assert len(result) == 1
    assert result[0]["expectation"] == "table_columns_exist_with_position"
    assert not result[0]["passed"]
    assert "extra" in result[0] and error in result[0]["extra"]["reason"]


@pytest.mark.parametrize("df", [pd.DataFrame([])])
def test_custom_check(df):
    """
    Test custom expectation
    """

    # => Create a custom expectation
    class CustomExpectation(ExpectationBase):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self.name = "custom_expectation"

    # Manager should pickup this new class
    mgr = ExpectationManagerBase()
    mgr.initialize()

    # Dont raise any error
    mgr.validate(df, {"expectation": "custom_expectation"})

    # Test position = 1
    result = mgr.validate(df, {"expectation": "custom_expectation", "params": {}})

    assert len(result) == 1
    assert result[0]["expectation"] == "custom_expectation"
    assert result[0]["passed"]



import os
import json
import pytest
import tempfile
import shutil
from enrichsdk.package.mock import *
from enrichsdk.policy import *
from enrichsdk.policy.sdk import *
from enrichsdk.lib import Context

invalid_policies = [
    (None, "Invalid specification format"),
    ({}, "Missing policy specification"),
    ({"args": {}}, "Missing policy specification"),
    ({"args": {"policies": {}}}, "Policy specification must be a list"),
    ({"args": {"policies": []}}, "Empty list of policies"),
    ({"args": {"policies": [{}]}}, "Invalid specification"),
    ({"args": {"policies": [{"principal": None}]}}, "Invalid specification"),
    ({"args": {"policies": [{"principal": ""}]}}, "Invalid specification"),
    ({"args": {"policies": [{"principal": "username"}]}}, "Invalid specification"),
    (
        {"args": {"policies": [{"principal": "username", "resource": None}]}},
        "Invalid specification",
    ),
    (
        {"args": {"policies": [{"principal": "username", "resource": "hello"}]}},
        "Invalid specification",
    ),
    (
        {
            "args": {
                "policies": [
                    {"principal": "username", "resource": "hello", "action": None}
                ]
            }
        },
        "Invalid specification",
    ),
    (
        {
            "args": {
                "policies": [
                    {
                        "principal": "username",
                        "resource": "hello",
                        "action": "hello",  # Invalid
                    }
                ]
            }
        },
        "Invalid action",
    ),
    (
        {
            "args": {
                "policies": [
                    {
                        "principal": "username",
                        "resource": "hello",
                        "action": "filter",
                        "args": None,
                    }
                ]
            }
        },
        "Invalid specification",
    ),
]

valid_policies = [
    {
        "args": {
            "policies": [
                {
                    "principal": "username:hello",
                    "resource": "confname::runname",
                    "action": "filter",
                    "args": {
                        "handler": "by_column_value",
                        "params": {"column": "storeid", "value": "attr::storeid"},
                    },
                }
            ]
        }
    }
]

invalid_principals = [
    ((None, None, None), "Invalid identifier specification"),
    (("", None, None), "Invalid identifier specification"),
    (([], None, None), "Invalid identifier specification"),
    (([1], None, None), "Invalid identifier specification"),
    (([None], None, None), "Invalid identifier specification"),
    ((["alpha"], None, None), "Invalid role specification"),
    ((["alpha"], "", None), "Invalid role specification"),
    ((["alpha"], {}, None), "Invalid role specification"),
    ((["alpha"], {1: 1}, None), "Invalid role specification"),
    ((["alpha"], {"a": 1}, None), "Invalid role specification"),
    ((["alpha"], {"a": ["B"]}, None), "Invalid attribute specification"),
    ((["alpha"], {"a": ["B"]}, ""), "Invalid attribute specification"),
    ((["alpha"], {"a": ["b"]}, {}), "Invalid attribute specification"),
    ((["alpha"], {"a": ["b"]}, {1: 1}), "Invalid attribute specification"),
]

invalid_resources = [None, ""]


resources = [BaseResource("alpha"), BaseResource("beta")]

principals = [
    BasePrincipal(
        identifiers=[
            "username:bob",
        ],
        roles={"marketing": ["admin"]},
        attrs={"storeid": 25},
    )
]

accesses = [
    {
        "principal": {
            "identifiers": ["username:hello"],
            "roles": {"marketing": ["admin"]},
            "attrs": {"storeid": 24},
        },
        "resource": {
            "name": "alpha",
        },
        "status": "deny",
    },
    {
        "principal": {
            "identifiers": ["username:hello"],
            "roles": {"marketing": ["admin"]},
            "attrs": {"storeid": 24},
        },
        "resource": {
            "name": "confname::runname",
        },
        "status": "filter",
    },
]


class TestPolicy(object):
    """
    Test the policy implementation
    """

    @classmethod
    def setup_class(cls):
        cls.enrich_root = tempfile.mkdtemp()
        context = Context({"ENRICH_ROOT": cls.enrich_root})
        cls.config = MockConfig(context.asdict())
        cls.state = MockState(config=cls.config)

    @classmethod
    def teardown_class(cls):
        if os.path.exists(cls.enrich_root):
            shutil.rmtree(cls.enrich_root)

    @pytest.mark.parametrize("prefix", invalid_resources)
    def test_invalid_resource(self, prefix):
        """
        Test invalid resources
        """
        with pytest.raises(Exception) as excinfo:
            b = BaseResource(prefix)

    @pytest.mark.parametrize("args,error", invalid_principals)
    def test_invalid_principal(self, args, error):
        """
        Test invalid principal
        """
        with pytest.raises(Exception) as excinfo:
            b = BasePrincipal(*args)

        assert error in str(excinfo)

    @pytest.mark.parametrize("conf,error", invalid_policies)
    def test_policy_invalid_configs(self, conf, error):
        """
        Test invalid policy configurations
        """
        with pytest.raises(Exception) as excinfo:
            policy = SimpleFilterEngine(config=self.config)
            policy.configure(conf)
            policy.validate()

        assert error in str(excinfo)

    def test_policy_valid_configs(self):
        """
        Test Valid configuration
        """
        # => Load the policy
        policy = SimpleFilterEngine(config=self.config)
        policy.configure(valid_policies[0])
        policy.validate()

    @pytest.mark.parametrize("params", accesses)
    def test_policy_execution(self, params):
        """
        Test implementation of the policy
        """

        principal = params["principal"]
        resource = params["resource"]
        status = params["status"]

        # Instantiate the policy engine
        conf = valid_policies[0]
        policy = SimpleFilterEngine(config=self.config)
        policy.configure(conf)

        # Create principal
        principal = BasePrincipal(**principal)

        # Construct resource
        resource = BaseResource(**resource)

        # Now apply the policy
        result = policy.apply(principal, resource)
        assert result["status"] == status



import pytest
from llmsdk.agents.prompter import LLMPromptAgent

# test cases with expected outputs
test_cases = [
    (
        "What is the first month of the year? Respond with only the month name in all caps.",
        "JANUARY"
    ),
    (
        "What day comes after Sunday? Respond with only the day name in lowercase.",
        "monday"
    ),
    (
        "Is an orange a fruit or a mammal? Respond with only the word fruit or animal in lowercase but with the first letter capitalized.",
        "Fruit"
    ),
]

class TestAgentPrompter:
    """
    Class containing test cases for the Prompter LLM Agent
    """

    # test methods
    @pytest.mark.parametrize("query,truth", test_cases)
    def test_agent_prompter(self, init_llm, query, truth):
        """
        method to create a Prompter Agent, send it a test prompt, and check the response
        """
        # test vars
        self.agentname          = init_llm["agentname"]
        self.platform           = init_llm["platform"]
        self.model              = init_llm["model"]
        self.embedding_model    = init_llm["embedding_model"]

        # create the agent
        agent = LLMPromptAgent(name=self.agentname,
                                 platform=self.platform,
                                 model=self.model,
                                 embedding_model=self.embedding_model)

        # setup the test prompt and expected response
        prompt = {
            "prompt": query
        }
        expected_response = truth

        # run the prompt and get the response
        response = agent.prompt(prompt)
        response = response.get('answer', "")

        # response should be JANUARY to pass
        assert response == expected_response



import os
import sys
import smtplib
import time
import mimetypes
import logging

from email import encoders
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from email.mime.audio import MIMEAudio
from email.mime.base import MIMEBase
from email.mime.image import MIMEImage

logger = logging.getLogger('app')

def attach_files(msg, attachments=[]):

    # https://docs.python.org/3.4/library/email-examples.html
    for path in attachments:

        if (not isinstance(path, str)) or (not os.path.isfile(path)):
            continue

        ctype, encoding = mimetypes.guess_type(path)
        if ctype is None or encoding is not None:
            ctype = "application/octet-stream"
        maintype, subtype = ctype.split("/", 1)
        if maintype == "text":
            with open(path) as fp:
                attach = MIMEText(fp.read(), _subtype=subtype)
        elif maintype == "image":
            with open(path, "rb") as fp:
                attach = MIMEImage(fp.read(), _subtype=subtype)
        elif maintype == "audio":
            with open(path, "rb") as fp:
                attach = MIMEAudio(fp.read(), _subtype=subtype)
        else:
            with open(path, "rb") as fp:
                attach = MIMEBase(maintype, subtype)
                attach.set_payload(fp.read())
            encoders.encode_base64(attach)

        filename = os.path.basename(path)
        attach.add_header("Content-Disposition", "attachment", filename=filename)
        msg.attach(attach)


def send_html_email(content, sender, receivers, subject, attachments=[], reply_to=[], cc=[], bcc=[], cred={}, retries=2):
    """
    Send HTML email to target audience
    """
    if cred is None:
        cred = {}

    retry_delay = 30 # seconds

    msg = MIMEMultipart()
    msg.attach(MIMEText(content, "html"))

    if isinstance(receivers, str):
        receivers = [receivers]

    msg["From"] = sender
    msg["To"] = ", ".join(receivers)
    if isinstance(reply_to, list) and len(reply_to) > 0:
        msg.add_header('reply-to', ",".join(reply_to))
    msg["Subject"] = subject
    if len(cc) > 0:
        msg['Cc'] = ",".join(cc)
    if len(bcc) > 0:
        msg['Bcc'] = ",".join(bcc)

    # Construct final list that includes everyone...
    to = receivers + cc + bcc

    # Add payload
    if isinstance(attachments, list) and len(attachments) > 0:
        attach_files(msg, attachments)

    server = cred.get("EMAIL_HOST", os.environ["EMAIL_HOST"])
    username = cred.get("EMAIL_HOST_USER", os.environ["EMAIL_HOST_USER"])
    password = cred.get("EMAIL_HOST_PASSWORD", os.environ["EMAIL_HOST_PASSWORD"])
    port = cred.get("EMAIL_PORT", os.environ.get("EMAIL_PORT", "587"))
    timeout = cred.get("EMAIL_TIMEOUT", os.environ.get("EMAIL_TIMEOUT", "30"))

    # Collect the body
    text = msg.as_string()

    msg = ""
    while retries > 0:
        try:
            smtp = smtplib.SMTP(server, port=int(port), timeout=int(timeout))
            smtp.starttls()
            smtp.login(username, password)
            smtp.sendmail(sender, to, text)
            return
        except Exception as e:
            msg += f"[Try {retries}] {e}\n"
            retries -= 1
            time.sleep(retry_delay)

    if len(msg) > 0:
        logger.error("Email sending process exhausted",
                     extra={
                         'data': msg
                     })
    raise Exception("Email sending process exhausted retries")



import os
import json
import copy
import shutil
import imp
import tempfile
import pytest

from enrichsdk.package.mock import MockConfig, MockState
from enrichsdk.lib import Context

basic_configs = [
    {"conf": None, "exceptions": {"config": lambda e: "valid dictionary" in str(e)}},
    {"conf": "", "exceptions": {"config": lambda e: "valid dictionary" in str(e)}},
    {"conf": {}, "exceptions": {"config": lambda e: "valid dictionary" in str(e)}},
    {
        "conf": {},
        "exceptions": {
            "config": lambda e: "valid dictionary" in str(e),
        },
        "check": [lambda instance: instance.is_enabled()],
    },
    {
        "conf": {"version": 2.0},
        "exceptions": {"config": lambda e: "Version mismatch" in str(e)},
    },
    {
        "conf": {"args": {}},
        "exceptions": {"validation": lambda e: "Invalid name" in str(e)},
    },
]

node_configs = [
    {
        "conf": {"name": "Hello", "args": {}},
        "exceptions": {"validation": lambda e: "Invalid node type" in str(e)},
    }
]

compute_configs = [
    {
        "conf": {"name": "Hello", "dependencies": None, "args": {}},
        "exceptions": {"validation": lambda e: "Dependencies" in str(e)},
    },
]


class TestClass:
    """
    Test whether I can instantiate a module
    """

    @classmethod
    def setup_class(cls):
        cls.enrich_root = tempfile.mkdtemp()
        context = Context({"ENRICH_ROOT": cls.enrich_root})
        cls.config = MockConfig(context.asdict())
        cls.state = MockState(config=cls.config)

    @classmethod
    def teardown_class(cls):
        if os.path.exists(cls.enrich_root):
            shutil.rmtree(cls.enrich_root)

    def load_module(self, dirname):
        """
        Helper routine..
        """
        root = os.path.join(os.path.dirname(__file__), "fixtures", "transforms")
        modpath = os.path.join(root, dirname)

        modname = os.path.basename(modpath)
        dirname = os.path.dirname(modpath)

        file_, path_, desc_ = imp.find_module(modname, [dirname])
        package = imp.load_module(modname, file_, path_, desc_)

        return package.provider

    @pytest.mark.parametrize(
        "dirname, configs",
        [
            ("node_test", basic_configs + node_configs),
            ("compute_test", basic_configs + compute_configs),
            ("transform_test", basic_configs + compute_configs),
            ("model_test", basic_configs + compute_configs),
            ("skin_test", basic_configs + compute_configs),
        ],
    )
    def test_config(self, dirname, configs):
        """
        Test raw node
        """

        provider = self.load_module(dirname)

        with pytest.raises(Exception) as excinfo:
            cls = provider()

        # This should not raise an exception
        for c in configs:
            # print(c)
            exceptions = c.get("exceptions", {})
            instance = provider(config=self.config)

            # Configuration check
            check = exceptions.get("config", None)
            if check is not None:
                with pytest.raises(Exception) as excinfo:
                    instance.configure(c["conf"])
                # print("Received error", str(excinfo))
                assert check(excinfo)
                continue
            else:
                instance.configure(c["conf"])

            # => Validation step...
            check = exceptions.get("validation", None)
            if check:
                with pytest.raises(Exception) as excinfo:
                    instance.validate("conf", self.state)
                # print(excinfo)
                # print("Invalid name" in str(excinfo))
                assert check(excinfo)
            else:
                instance.validate("conf", self.state)

            # Run the checks on this node
            for check in c.get("checks", []):
                assert check(instance)

    def test_transform(self):
        """
        Test transform
        """
        provider = self.load_module("transform_test")
        conf = {"dependencies": None, "args": {}}

        t = provider(config=self.config)
        t.configure(conf)

        assert t.is_transform()



from __future__ import print_function
from sys import getsizeof, stderr
from itertools import chain
from collections import deque
try:
    from reprlib import repr
except ImportError:
    pass

#=> Found this on stackoverflow
def total_size(o, handlers={}, verbose=False):
    """Returns the approximate memory footprint an object and all of its
    contents.

    Automatically finds the contents of the following builtin
    containers and their subclasses: tuple, list, deque, dict, set and
    frozenset.  To search other containers, add handlers to iterate
    over their contents:

        handlers = {SomeContainerClass: iter,
                    OtherContainerClass: OtherContainerClass.get_elements}

    """
    dict_handler = lambda d: chain.from_iterable(d.items())
    all_handlers = {
        tuple: iter,
        list: iter,
        deque: iter,
        dict: dict_handler,
        set: iter,
        frozenset: iter,
    }
    all_handlers.update(handlers)     # user handlers take precedence
    seen = set()                      # track which object id's have already been seen
    default_size = getsizeof(0)       # estimate sizeof object without __sizeof__

    def sizeof(o):
        if id(o) in seen:       # do not double count the same object
            return 0
        seen.add(id(o))
        s = getsizeof(o, default_size)

        if verbose:
            print(s, type(o), repr(o), file=stderr)

        for typ, handler in all_handlers.items():
            if isinstance(o, typ):
                s += sum(map(sizeof, handler(o)))
                break
        return s

    return sizeof(o)



from .agent_events import *

from .basellmrag import *
from .datagpt import *
from .docgen import *
from .databank import *
from .mltclassifier import *
from .multistep import *
from .prompter import *
from .qna import *
from .querymapper import *
from .sadl import *



import os
import json
import time
import string
import hashlib

from langchain_openai import ChatOpenAI
from langchain_core.messages import HumanMessage, SystemMessage
from langchain_community.callbacks.manager import get_openai_callback

from . import agent_events
from llmsdk.lib import defaults
from llmsdk.lib.enablers import AgentEnablersMixin
from llmsdk.lib import SafeEncoder
from llmsdk.services.log import *

__all__ = ['LLMPromptAgent']

class LLMPromptAgent(AgentEnablersMixin):
    """
    A geenric agent to run a prompt
    """

    def __init__(self,
                 name,
                 cred={},
                 platform=defaults.LLM_PLATFORM,
                 model=defaults.LLM_MODEL,
                 embedding_model=defaults.LLM_EMBEDDING_MODEL):
        """
        init the bot
        name: name of the bot
        cred: credentials object
        platform: name of the platform backend to use
                default to openai platform for now
                will be extended in the future to suuport other platforms
        model: model name
        """

        start_time = time.time()

        # logging
        self.logger = get_logger()

        # defaults
        self.metadata       = {}
        self.max_llm_tokens = 4000

        # name
        self.agent_name = name
        self.agent_type = "prompter"
        self.agent_id   = self._create_id(f"{self.agent_name}_{start_time}")

        # creds
        self.cred = cred
        # LLM params
        self.platform           = platform
        self.model              = model
        self.embedding_model    = embedding_model

        # init the llm and embeddings objects
        self.llm, self.embeddings = self._get_llm_objs(platform=self.platform,
                                                          model=self.model,
                                                          embedding_model=self.embedding_model,
                                                          cred=self.cred)

        # agent defaults
        self.persona_prompt = "You are a highly advanced AI agent capable of solving tasks that you are presented with."

        self.errs = {
            "human_msg": "No prompt available, cannot call LLM",
        }

        # note metadata for this agent
        self.metadata = {
            "agent": {
                "name": self.agent_name,
                "name": self.agent_type,
                "platform": self.platform,
                "model": self.model,
                "embedding_model": self.embedding_model,
            },
            "events": []
        }
        # log that the agent is ready
        duration = time.time() - start_time
        event = self._log_event(agent_events._EVNT_READY, duration)

    ## interfaces

    def run_prompt(self, sys_msg, human_msg):
        """
        generate a prompt for querying the LLM
        """
        # construct the prompt template
        messages = [
            SystemMessage(content=sys_msg),
            HumanMessage(content=human_msg),
        ]

        # call the LLM
        response = self.llm.invoke(messages)

        return response

    def prompt(self, prompt):
        """
        run a query using llm to answer questions about some text data
        this is useful when looking for answers about some context but without doing RAG
        """
        # construct the prompt
        sys_msg = prompt.get("persona", self.persona_prompt)
        human_msg = prompt.get("prompt")

        if human_msg == None:
            result = { "answer": self.errs["human_msg"] }

        else:
            # run prompt
            response = self.run_prompt(sys_msg, human_msg)
            result = { "answer": response.content }

        return result



import pandas as pd

class BaseDataScorer(object):
    """
    Class to process data and produce some usable output
    This class is generic enough to handle arbitrary pandas dataframes
    """

    def __init__(self):
        """
        init the class
        """
        return

    ##
    ## internal methods
    ##

    def _spec_filter(self, df_orig, filters):
        """
        Apply a set of filters to a form5500 features dataframe
        """

        steps = []
        df = df_orig.copy()

        for fltr in filters:

            name = fltr.get("name", "")

            # get the params needed
            col = fltr.get("column")
            op = fltr.get("op")
            val = fltr.get("value")

            # check for correctness
            if (col is None) or (op is None) or (val is None):
                continue

            # perform the operation
            if op == "between":
                if len(val) != 2:
                    continue
                df = df[(df[col]>=val[0]) & (df[col]<=val[1])]
                step = f"{col} between {val[0]} and {val[1]}"
                steps.append(step)

            if op == "contains":
                if len(val) == 0:
                    continue
                df = df[df[col].str.contains('|'.join(val), na=False, case=False)]
                step = f"{col} containing any of {val}"
                steps.append(step)

            if op == "exclude":
                if len(val) == 0:
                    continue
                df = df[~df[col].str.contains('|'.join(val), na=False, case=False)]
                step = f"{col} to exclude all of {val}"
                steps.append(step)

            if op == "gte":
                df = df[df[col]>=val]
                step = f"{col} >= {val}"
                steps.append(step)

            if op == "gt":
                df = df[df[col]>val]
                step = f"{col} > {val}"
                steps.append(step)

            if op == "lte":
                df = df[df[col]<=val]
                step = f"{col} <= {val}"
                steps.append(step)

            if op == "lt":
                df = df[df[col]<val]
                step = f"{col} < {val}"
                steps.append(step)

        # generate a note with all steps
        steps = "-> " + "\n-> ".join(steps)

        return df, steps

    def _spec_focus(self, df_orig, focus):
        """
        Apply a set of focus steps to a form5500 features dataframe
        """

        steps = []
        df = df_orig.copy()

        for fcs in focus:

            name = fcs.get("name", "")

            # get the params needed
            col = fcs.get("column")
            order = fcs.get("order")
            weight = fcs.get("weight")

            # check for correctness
            if (col is None) or (order is None) or (weight is None):
                continue

            # create a normalized score (0 - 1) from the column values
            fcs_col_name = f"fcs::{col}"
            min_val = min(df[col])
            max_val = max(df[col])
            df[fcs_col_name] = (df[col] - min_val) / (max_val - min_val)

            # flip the score if the ordering is lowest-first
            if order == "asc":
                df[fcs_col_name] = 1 - df[fcs_col_name]

            # apply the weight
            df[fcs_col_name] = weight * df[fcs_col_name]

            # note the step
            step = f"Computed normalized score for {col} in {order.upper()} order and weighted by {weight}"
            steps.append(step)

        # generate the consolidated score
        cols = [c for c in df.columns if 'fcs::' in c]
        df['fcs::score'] = 0
        for c in cols:
            df['fcs::score'] = df['fcs::score'] + df[c]

        # generate a note with all steps
        steps = "-> " + "\n-> ".join(steps)

        return df, steps

    def _spec_sort(self, df_orig, sort):
        """
        Apply a sort ordering to a form5500 features dataframe
        """

        steps = []
        sort_cols = []
        sort_order = []
        df = df_orig.copy()

        for srt in sort:

            name = srt.get("name", "")

            # get the params needed
            col = srt.get("column")
            order = srt.get("order")

            # construct the sort order
            sort_cols.append(col)
            sort_order.append(True if order=="asc" else False)

            # note the step
            step = f"Sorted by {col} in {order.upper()} order"
            steps.append(step)


        # note the step for the sort by fcs::score
        step = f"Sorted by the Consolidated Weighted Normalized Score in DESC order"
        steps.append(step)

        # apply the sort order
        sort_cols.append('fcs::score')
        sort_order.append(False)
        df = df.sort_values(sort_cols, ascending=sort_order)

        # re-order columns
        first_cols = ['ACK_ID'] + sort_cols
        cols = [c for c in df.columns if c not in first_cols]
        df = df[first_cols + cols]

        # generate a note with all steps
        steps = "-> " + "\n-> ".join(steps)

        return df, steps

    def _get_steps_commentary(self, filter_steps, focus_steps, sort_steps):
        """
        Construct commentary on the steps taken
        """

        all_steps = ""

        if len(filter_steps) > 0:
            all_steps += "The Form5500s were filtered using these rules:" + "\n"
            all_steps += f"{filter_steps}" + "\n\n"
        if len(focus_steps) > 0:
            all_steps += "The Form5500s were scored using these rules:" + "\n"
            all_steps += f"{focus_steps}" + "\n"
            all_steps += "And a Consolidated Weighted Normalized Score was computed" + "\n\n"
        if len(sort_steps) > 0:
            all_steps += "The Form5500s were sorted in this order:" + "\n"
            all_steps += f"{sort_steps}" + "\n\n"

        return all_steps

    ##
    ## interfaces
    ##

    def process(self, df, spec):
        """
        Take a form5500 features dataframe and a spec
        and apply the spec to the dataframe
        """

        # check if there are columns to fill NaNs with 0
        fill_zero_cols = spec.get("focus", [])
        for column in fill_zero_cols:
            col = column.get("column")
            if col != None:
                df[col] = df[col].fillna(0)

        # apply the filters
        filter_steps = ""
        filters = spec.get("filter", [])
        if len(filters) > 0:
            df, filter_steps = self._spec_filter(df, filters)

        # apply the focus
        focus_steps = ""
        focus = spec.get("focus", [])
        if len(focus) > 0:
            df, focus_steps = self._spec_focus(df, focus)

        # apply the sort
        sort_steps = ""
        sort = spec.get("sort", [])
        if len(sort) > 0:
            df, sort_steps = self._spec_sort(df, sort)

        # construct the notes commentary
        all_steps = self._get_steps_commentary(filter_steps, focus_steps, sort_steps)

        return df, all_steps
